#!/pro/bin/perl

package PROCURA::DBD;

use strict;
use warnings;

use vars qw($VERSION);

$VERSION = 1.14;

=head1 NAME

DBD - Generic perl DBI/DBD API for Unify, Oracle, and MySQL databases

=head1 SYNOPSIS

    use PROCURA::DBD;

    DBDlogon ();			# Logon to current database - read-only
    $dbh = DBDlogon (1);		# same, read-write, return database handle
    DBDlogon ({ ReadOnly   => 1 });	# Try read-only mode
    DBDlogon ({ DBDverbose => 1 });	# specify options

    my ($d_in, $d_end) =
	getrow ("select d_in, d_end from table where c_table = 1");

    my @tbl = tables ([ Schema ]);      # return the *unquoted* list of tables
                                        # see also /pro/lib/prod_tbl.seq
    my $stp = prepar (
	"select  d_in, d_end",
	"from    table",
	"where   c_table = ?");		# prepare statement for later execution
    my $sth = prepex (
	"select  c_table, table",
	"from    table",
	"where   c_table > 1");		# prepare and execute immediate
    my $sth = prepex (
	"select  c_table,   table",
	       \$c_table, \$table,
	"from    table",
	"where   c_table > 1");		# prepex and bind_columns at once

    my @desc = describe ("table");	# Get table description as an array of
					# hashes. $desc[0]{NAME} is the name of
					# the first field
    my @desc = describe ("table", "L");	# .. *with* link info in $desc[0]{LINK}

    DBDlogoff ();			# Using DBD, don't use $dbh->disconnect

=head1 EXAMPLES

    # using $stp and $sth from SYNOPSYS:
    while (my ($c_table, $table) = $sth->fetchrow_array) {
	$stp->execute ($c_table);
	my ($d_in, $d_end) = $stp->fetchrow_array;
	# :
	}

    $stp->execute (3);
    $stp->bind_columns (\$d_in, \$d_end);
    while ($stp->fetch) {
	print "Period: $d_in .. $d_end\n";
	}
    $stp->finish;

    my @c_naam = sort { $a <=> $b }
	getrows ("select c_naam from naam where diac = 1");
    my $sta = prepar (
	"select unique c_straat",
	"from   naw",
	"where  c_gem_deel = ?");
    my @c_straat = $sta->getrows (2);

    my $stb = prepex (
	"select c_straat,   hnr,",
	     \($c_straat,  $hnr),
	"       c_plaats,   pc",
	      \$c_plaats, \$pc,
	"from   naw");
    while ($stb->fetch) {
	print s_straat ($c_straat), $hnr, ...
	}

=head1 DESCRIPTION

=over 4

=item

=item [ C<$dbh> = ] B<DBDlogon (>[C<1>], [C<{ attributes }>]B<)>

Make a connection to the current database. This call is optional, if the first
DBD related call is one of the routines exported by PROCURA::DBD, which does
automatically call DBDlogon if not yet connected.

B<DBDlogon ()> returns the database handle, see DBD::Unify and DBI to see what
you can do with that. There's two optional parameters. One determines if the
Database is opened for read-only (0), which is the default, or read-write (1).
With the current implementation, this is only used for Oracle databases.

The scond optional parameter, passes additional attributes to be used when
connection to the database. These will be added (or overrule) to the default:

  { RaiseError         => 1,
    PrintError         => 1,
    AutoCommit         => 0,  # if supported
    ChopBlanks         => 1,
    ShowErrorStatement => 1,
    FetchHashKeyName   => "NAME_lc",
    }

The Unify connect also passes { uni_scanlevel => 7 }

Unify uses B<$USCHEMA>, Oracle uses B<$DBUSER> for logon data.

=item B<DBDlogoff ()>

Disconnect the current database connection. This call should only be used if
another connection will be made after the call within the same program. The
disconnection will be done automatically in the C<END {}> of the module,
enabling other C<END {}> blocks to terminate handles in the correct sequence.

=item B<prepar ()> / B<prepex ()>

C<prepar ()> takes two types of arguments, intermixed. All simple scalars
will be joined together to a SQL statement, which in turn is passed to
the DBI C<prepare> method. All other arguments, if any, should be
references and are passed to DBI method C<bind_columns> for the just prepared
statement handle.

C<prepex ()> first calls C<prepar ()> with all its arguments and then calls
DBI method C<execute> for that statement handle.

=item B<tables ()>

As of DBI-1.36 a very bad design change caused C<$dbh->tables> method to return
all table names quoted. This function call has two advantages over the DBI
method: 1. You don't need the database handle; 2. It will return the table
names unquoted, as it should have been, and always was till DBI 1.35

C<tables ()> takes an optional C<SCHEMA> argument, so it'll return only the
tables from C<SCHEMA>.

=back

=cut

use DBI 1.35;

use Carp;
use Exporter;
use vars qw(@ISA @EXPORT);
@ISA = ("Exporter", scalar caller);
@EXPORT = qw(
    DBDlogon DBDlogoff
    tables
    describe
    prepar prepex
    getrow getrows
    local_sql sql_into
    );

umask 0;

my $dbh;		# Everyone else uses prepex () or do ()
my $dbh_st = "";	# to be replaced with $dbh->{Statement} in future DBI
			# which will also support ->do () et all

$SIG{__DIE__} = sub {
    $_[0] =~ m/^DBD::/ or return;
    $dbh && ($dbh->err || $dbh->state) or return;

    printf STDERR "DBI:DBD Failure from line %d in %s\n", (caller (1))[2, 1];
    $dbh->err    and print STDERR "    err:\t",    $dbh->err,    "\n";
    $dbh->errstr and print STDERR "    errstr:\t", $dbh->errstr, "\n";
    $dbh->state  and print STDERR "    state:\t",  $dbh->state,  "\n";
    if ($DBI::lasth) {
	my $s = $DBI::lasth->{Type} eq "st" ? $DBI::lasth->{Statement} : $dbh_st;
	print STDERR "    ", (join "\n    " =>
	    "--------",
	    split (m/\n/ => $s),
	    "--------"), "\n";
	}
    die "";
    }; # __DIE__

END {
    $dbh and DBDlogoff ();
    }

### SQL utils #################################################################

sub DBDlogon (;$$)	# Default Read-Only
{
    $dbh and return $dbh;

    my $wr   = 0;
    my %attr = (
	RaiseError         => 1,
	PrintError         => 1,
	AutoCommit         => 0,
	ChopBlanks         => 1,
	ShowErrorStatement => 1,
	);
    @_ && !ref $_[0] and $wr = shift;	# Depricated
    if (@_ &&  ref $_[0]) {
	my $r = shift;
	foreach my $attr (keys %$r) { $attr{$attr} = $r->{$attr} }
	}

    if (exists $attr{ReadOnly}) {
	$wr = !$attr{ReadOnly};
	delete $attr{ReadOnly};
	}

    if (exists $ENV{DBPATH} && -f "$ENV{DBPATH}/file.dbs") {
	exists $attr{ScanLevel} and $attr{uni_scanlevel} = delete $attr{ScanLevel};
	exists $attr{uni_scanlevel} or $attr{uni_scanlevel} = 7;
	$dbh = DBI->connect ("DBI:Unify:", "", $ENV{USCHEMA}, \%attr) or
	    croak "connect: $!";
	}
    elsif (exists $ENV{ORACLE_HOME} && -d $ENV{ORACLE_HOME}) {
	$attr{FetchHashKeyName} = "NAME_lc";
	my $up = $ENV{DBUSER} || $ENV{ORACLE_USERID} || uc "PRO$ENV{PROD}/PRO$ENV{PROD}";
	my ($dbu, $dbp) = split m:/: => $up;
	defined $dbp && $dbp =~ m/\S/ or $dbp = $dbu;
	$dbh = DBI->connect ("DBI:Oracle:", $dbu, $dbp, \%attr) or
	    croak "connect: $!";
	# $wr or $dbh->do ("set transaction read only");
	}
    elsif (exists $ENV{PGDATA}) {
	$attr{FetchHashKeyName} = "NAME_lc";
	(my $db = $ENV{DBI_DSN} // getpwuid $<) =~ s/.*=//;
	my ($dbu, $dbp) = map { $ENV{"DBI_$_"} // undef } qw( USER PASS );
	$dbh = DBI->connect ("DBI:Pg:dbname=$db", $dbu, $dbp, \%attr) or
	    croak "connect: $!";
	}
    else {
	my $db = exists $ENV{MYSQLDB} ? $ENV{MYSQLDB} : "test";
	delete $attr{AutoCommit};	# MySQL still croaks on this one
	$dbh = DBI->connect ("DBI:mysql:database=$db", $ENV{LOGNAME}, undef, \%attr) or
	    croak "connect: $!";
	}
    $dbh;
    } # DBDlogon

sub DBDlogoff ()
{
    $dbh and $dbh->disconnect;
    $dbh = undef;	# No guarantee anyone else still holds the handle
    } # DBDlogoff

# prepar () / prepex () serve two purposes:
# 1. Ease passed statement to allow statement lines in array
# 2. Hide $dbh to the outside world
# 3. Enable immediate column binding
# 4. Combine the ever occuring execute method after the prepare
# We do not have to check anything, since RaiseError is on
sub prepar (@)
{
    $dbh or DBDlogon ();
    my (@st, @bc);
    for (@_) { ref $_ ? push @bc, $_ : push @st, $_ }
    $dbh_st = join "\n", @st;
    my $sth = $dbh->prepare ($dbh_st);
    if (@bc) {
	my @c = caller (1);
	(my $sub = $c[3] // "") =~ s/.*:://;
	$sub eq "prepex" && $c[1] eq __FILE__ and return ($sth, @bc);
	$sth->bind_columns (@bc);
	}
    $sth;
    } # prepar

sub prepex (@)
{
    my ($sth, @bc) = prepar (@_);
    $sth->execute;
    @bc and $sth->bind_columns (@bc);
    $sth;
    } # prepex

sub tables (;$$$$)
{
    my @args = @_;
    @args == 0 and @args = (undef, undef, undef, undef);
    @args == 1 and @args = (undef, @args, undef, undef);
    @args == 2 and @args = (undef, @args, undef);
    @args == 3 and @args = (undef, @args);
    $dbh or DBDlogon ();
    map { s/"//g; $_ } $dbh->tables (@args);
    } # tables

sub getrow (@)
{
    $dbh or DBDlogon ();
    $dbh_st = join "\n", @_;
    $dbh->selectrow_array ($dbh_st);
    } # getrow

sub getrows (@)
{
    my $sth = prepex (@_);
    $sth->getrows;
    } # getrows

sub describe ($;$)
{
    $dbh or DBDlogon ();
    my ($schema, $table) = split m/\./, $_[0];
    $schema && !$table and ($schema, $table) = (undef, $schema);
    my %li;
    if ($_[1]) {
	my $li;
	if (exists $ENV{ORACLE_SID}) {
	    my $stl = $dbh->foreign_key_info (undef, undef, undef, undef, $schema, uc $table);
	    while ($stl and $li = $stl->fetchrow_hashref) {
		$li{lc $li->{fk_column_name}} = join ".",
		    $li->{uk_table_schem},
		    map { lc } @{$li}{qw( uk_table_name uk_column_name )};
		}
	    }
	elsif (exists $ENV{USCHEMA}) {
	    my $stl = DBD::Unify::db::link_info ($dbh, undef, $schema, $table, undef);
	    while ($stl and $li = $stl->fetchrow_hashref) {
		$li{$li->{REFERENCING_COLUMN}} = join ".",
		    @{$li}{qw(REFERENCED_OWNER REFERENCED_TABLE REFERENCED_COLUMN)};
		}
	    }
	}
    my $sth = prepex ("select * from $_[0] where 0 = 1");
    my @desc;
    my @name = @{$sth->{NAME}};
    foreach my $i (0 .. $#name) {
	foreach my $col (qw( NAME NAME_lc NAME_uc
			     TYPE PRECISION SCALE NULLABLE )) {
	    exists $sth->{$col} and
		$desc[$i]{$col} = $sth->{$col}[$i];
	    }
	if (exists $ENV{PGDATA}) {
	    if (my $x = $dbh->column_info ("", $schema, $table, $sth->{NAME}[$i])) {
		my $h = $x->fetchrow_hashref;
		@{$desc[$i]}{qw( NULLABLE PRECISION SCALE )} =
		    ($h->{NULLABLE}, reverse split m/,/ => $h->{COLUMN_SIZE});
		}
	    }
	$desc[$i]{LINK} = $li{$sth->{NAME_lc}[$i]} // undef;
	}
    @desc;
    } # describe

### ##########################################################################

*local_sql = \&getrows;	# Backward compatability, to be removed

sub sql_into ($@)
{
    my $file = shift;

    open my $out, ">$file" or croak "$file: $!";

    my $sth = prepex (@_);
    while (my @f = $sth->fetchrow_array) {
	print $out join ("|" => @f), "\n";
	}
    $sth->finish;
    close $out;
    } # sql_into

### ###########################################################################

# These three provide methods that prevent NULL values in the database
# Oracle translates empty VARCHAR2 values to NULL :[[[

sub DBI::st::insert ($@)
{
    my $sth = shift;
    $sth->execute (map { defined ($_) && $_ eq "" ? " " : $_ } @_);
    } # insert

sub DBI::st::update ($@)
{
    my $sth = shift;
    $sth->execute (map { defined ($_) && $_ eq "" ? " " : $_ } @_);
    } # update

sub DBI::st::getrows ($@)
{
    my $sth = shift;
    my @r = ();
    $sth->execute (@_);
    while (my @f = $sth->fetchrow_array) {
	push @r, @f == 1 ? $f[0] : [ @f ];
	}
    @r;
    } # getrows
