package DBI::db;

# PLEASE write docs!
# Please write tests

$PROCURA::DBD_create::VERSION = 0.04;

=head NAME

DBD_create - Extension to the PROCURA::DBD module to do DD updates

=head1 SYNOPSYS

 use PROCURA::DBD;
 use PROCURA::DBD_create;

 my $dbh = DBDlogon (1);

 $dbh->drop_table ("table");

 $dbh->create_table ("tmp (key numeric (4) not null primary key)");

 $dbh->create_table (qq;
   create table tbl (
     c_usr   numeric (9) not null,  -- usr.c_usr
     d_in    numeric (8) not null,
     pid     numeric (9),
     c_appl  numeric (4),           -- appl.c_appl
     aant    char   (20),
     primary key (c_usr, d_in),
     );, count => 10000);

 $dbh->create_hash ("tmp.key");
 $dbh->create_hash ("tbl", "c_usr", "d_in", ...)
 $dbh->create_hash ("tbl", [ "c_usr", "d_in", ... ]);
 
 $dbh->create_link ("tbl.c_appl" => "appl.c_appl");
 $dbh->create_link (qw( tbl c_appl appl c_appl ));
 $dbh->create_link ("foo", [ "c_afn", "d_in" ] =>
                    "tbl", [ "c_usr", "d_in" ]);

 $dbh->create_index ("tmp.key");
 $dbh->create_index ("tbl", "c_usr", "d_in", ...);
 $dbh->create_index ("tbl", [ "c_usr", "d_in", ... ]);
 $dbh->create_index ("tbl.field", unique => 1, verbose => 1);
 $dbh->create_index ("tbl", [ "c_usr", "d_in" ], unique => 1);

 $dbh->commit;

=head1 DESCRIPTION

=over 4

=item B<drop_table (>C<$table>B<)>

Drop table C<$table>. In Oracle, it adds C<cascade constraints>,
something done automatically in Unify.

=item B<create_table (>C<table_def> [, C<options>]B<)>

Create a database table, given a genericarized interface. Some
characteristics of this method:

=over 2

=item -

Leading C<create table> is optional, since the method name is clear enough;

=item -

Comments and empty lines are stripped;

=item -

Unify field types are the prefered notation, as they are automatically
converted to Oracle: C<numeric> and C<number> are dealt with as being
the same, as are C<string>, C<char>, and C<varchar2>.

Unify types C<amount> and C<huge amount> are converted to Oracle's
multi-arg C<number> types;

=item -

Primary key defenitions are automatically converted to Oracle's way
of alter table constraints;

=item -

Table size can be passed as option C<count =E<gt> number>, or
C<segment =E<gt> size>. The first is prefered;

=back

=item B<create_hash ($;@)>

Since we only build hashes on primary keys, this is a no-op in
Oracle, since it does not allow indices on primary keys;

=item B<create_link ($$;$$$$$)>

Create a link, or reference to a foreign key, from fields in one
table to the primary key field(s) in another;

=item B<create_index ($;@)>

Create an (unique) index on one or more fields in a table.

=back

Except for C<drop_table ()>, all methods accept C<verbose =E<gt> 1>
as last arguments to show what is executed.

None of these methods do a commit. That's up to the calling program.

=cut

use base "DBI";

sub _roundup ($)
{
    my $ref = shift;
    $$ref <= 50 and return 50;
    my $x = 64;
    $x <<= 1 while $$ref > $x;
    $$ref = $x;
    } # _roundup

sub DBI::db::drop_table ($$)
{
    my ($dbh, $tbl) = @_;
    my $ddd = "drop table $tbl";
    $ENV{DBTYPE} eq "ORACLE" and $ddd .= " cascade constraints";
    $dbh->do ($ddd);
    } # drop_table

sub DBI::db::create_table ($$;@)
{
    my ($dbh, $ddd, %options) = @_;
    my ($bytes, $table, @cnf, $pk) = (0);

    # "create table" optional. The method name tells enough :)
    $ddd =~ s/^\s*(?:create\s+table\s+)?(\S+)/create table $1/i or return;
    $table = $1;

    # Strip comments and empty lines
    $ddd = join "\n",
	grep m/\S/, map { s/--.*$//;  s/\s+$//; $_ } split m/\n/, $ddd;

    # be leanient against a comma after the last field
    $ddd =~ s/,([\s\n]*\))/$1/g;

    $options{verbose}   //= 0;
    unless (exists $options{segment} or exists $options{count}) {
	$options{count} //= 120;
	$options{count}   < 120 and $options{count} = 120;
	}
    $options{count} = int $options{count};

    $ddd =~ s/\b string \b\s* \(/char    \(/gx;
    if ($ENV{DBTYPE} eq "ORACLE") {
	my $ts = (split m:/:, $ENV{DBUSER}||$ENV{ORACLE_USERID})[0];
	$ddd =~ s/ \b numeric \b \s* \(/number   (/gx;
	$ddd =~ s/ \b char    \b \s* \(/varchar2 (/gx;

	# Detecting amounts:
	# - amount (7)  configuration (display (9, 2))
	# - huge amount configuration (display (18, 2))
	$ddd =~ s/ \b amount \s* \(\s* (\d+) \s*\)
		    (?: [\s\n]* configuration \s*\(
			[\s\n]* [^)]*? display \s* \( (\d+) \s*,\s* (\d+) \s*\) [^)]* \)
			\s*
		    )?  /number   ($1-$2-$3)/gx;
	$ddd =~ s/ \b huge amount \s*
		    (?:     configuration \s*\(
			[\s\n]* [^)]*? display \s* \( (\d+) \s*,\s* (\d+) \s*\) [^)]* \)
			[\s\n]*
		    )?  /number   (1-$1-$2)/gx;
	$ddd =~ s/(\d+)-(\d+)-(\d+)/sprintf"%d, %d", $2 > $1 ? $2 - 1 : $1, $3/ge;

	$bytes += $1 while $ddd =~ m/\(\s*(\d+)/g;

	$options{initial} //= 10 * int ($options{count} * $bytes / 10240);
	$options{initial}   < 50 and $options{initial} = 50;
	$options{"next"}  //= int ($options{initial} / 4);
	$options{"next"}    < 50 and $options{"next"} = 50;
	@cnf = map  { "$_ $options{$_}K"  }
	       grep { exists $options{$_} && _roundup (\$options{$_}) }
	       qw( initial next );
	$options{pctincrease} //= 0;
	push @cnf, "pctincrease $options{pctincrease}";
	$ddd =~ s/\)\s*;?\s*$/) tablespace $ts storage (@cnf)/;
	if ( $ddd =~ s/,[\s\n]*primary\s+(key)\s*\(([^)]+)\)// or
	     $ddd =~ s/^(\s*(\S+)\b.*)\s+primary key\b/$1/m) {
	    for (qw( initial next )) {
		use integer;
		$options{$_} /= 4;
		_roundup (\$options{$_});
		}
	    $pk = join "\n",
		"alter table $ts.$table",
		"    add constraint H_$table",
		"	primary key ($2)",
		"	using index tablespace $ts",
		"	storage (initial $options{initial}K next $options{'next'}K pctincrease 0)";
	    }
	}
    elsif (exists $ENV{PGDATA}) {
	# No option support (yet)
	}
    else {
	exists $options{count} and
	    push @cnf, "estimated count is $options{count}";
	exists $options{segment} and
	    push @cnf, "segment is $options{segment}";
	exists $options{description} and
	    push @cnf, "description '$options{description}'";
	if (@cnf) {
	    my $cnf = join ",\n\t\t", @cnf;
	    $ddd =~ s/\(/\(\n\tconfiguration ($cnf),/;
	    }
	}
    for ($ddd, $pk) {
	$_ or next;
	$options{verbose} and print STDERR "$_;\n";
	$dbh->do ($_);
	}
    } # create table

# $dbh->create_hash ("table.field")
# $dbh->create_hash ("table", "field 1", "field 2", ...)
# $dbh->create_hash ("table", [ "field 1", "field 2", ... ])
sub DBI::db::create_hash ($$;@)
{
    my ($dbh, $tbl, @fld, $hsh) = @_;
    my $opt_v = 0;
    if (@fld > 1 and $fld[-1] =~ m/^\d+$/ && $fld[-2] eq "verbose") {
	$opt_v = 0 + pop @fld;
	pop @fld;
	}
    $tbl =~ m/^(\S+)\.(\S+)$/ && @fld == 0 and
	($tbl, @fld) = ($1, $2);
    @fld == 1 && ref $fld[0] and @fld = @{$fld[0]};
    my $nm = "H_$tbl";
    length ($nm) > 30 and substr ($nm, 30) = "";
    if ($ENV{DBTYPE} eq "ORACLE") {
	# Hashes are build on primary keys, Oracle
	# does not allow this
	$hsh = "";
	}
    else {
	local $" = ", ";
	$hsh = "create unique hash index $nm on $tbl (@fld)";
	}
    $hsh or return;
    $opt_v and print STDERR $hsh, "\n";
    $dbh->do ($hsh);
    } # create_hash

# $dbh->create_link (qw( ll_adr c_ll leerling c_ll ));
# $dbh->create_link ("bewon", "c_straat", "straat", "c_straat");
# $dbh->create_link ("vb.c_plaats" => "plaats.c_plaats");
# $dbh->create_link ("afn", [ "c_afn", "d_in" ] => "afnemer", [ "afn", "d_afn" ]);
sub DBI::db::create_link ($$$;$$$$$)
{
    my $dbh = shift;
    my $opt_v = 0;
    if (@_ > 1 and $_[-1] =~ m/^\d+$/ && $_[-2] eq "verbose") {
	$opt_v = 0 + pop @_;
	pop @_;
	}
    my ($tt, $tf) = $_[0] =~ m/^(\S+)\.(\S+)$/ ? ($1, $2, shift) : (splice @_, 0, 2);
    my ($rt, $rf) = $_[0] =~ m/^(\S+)\.(\S+)$/ ? ($1, $2, shift) : (splice @_, 0, 2);
    ref $rf or $rf = [ $rf ];
    ref $tf or $tf = [ $tf ];
    my $ln = shift || "L_${tt}_$tf->[0]_${rt}";
    $ln =~ s/_PRO(\w\w\w)\./_\U$1\_/i;
    length ($ln) > 30 and substr ($ln, 30) = "";
    local $" = ", ";
    if ($ENV{DBTYPE} eq "ORACLE") {
	my $rs = (split m:/:, $ENV{DBUSER}||$ENV{ORACLE_USERID})[0];
	my $ts = $rt =~ s/^(\w+)\.(\S+)/$2/ ? $1 : $rs;
	$ln = qq;
	    alter table $rs.$tt add constraint $ln foreign key
		(@$tf) references $ts.$rt (@$rf);
	}
    else {
	$ln = qq;
	    create link index $ln on $tt (@$tf)
		references $rt (@$rf);
	}
    $opt_v and print STDERR $ln, "\n";
    $dbh->do ($ln);
    } # create_link

# $dbh->create_index ("table.field")
# $dbh->create_index ("table", "field 1", "field 2", ...)
# $dbh->create_index ("table", [ "field 1", "field 2", ... ])
# $dbh->create_index ("table.field", unique => 1, verbose => 1)
sub DBI::db::create_index ($$;@)
{
    my ($dbh, $tbl, @fld, $idx) = @_;
    my ($opt_v, $unique) = (0, "");
    if (@fld > 1 and $fld[-1] =~ m/^\d+$/ && $fld[-2] eq "verbose") {
	$opt_v = 0 + pop @fld;
	pop @fld;
	}
    if (@fld > 1 and $fld[-1] =~ m/^\d+$/ && $fld[-2] eq "unique") {
	$unique = pop @fld ? " unique" : "";
	pop @fld;
	}
    $tbl =~ m/^(\S+)\.(\S+)$/ && @fld == 0 and
	($tbl, @fld) = ($1, $2);
    @fld == 1 && ref $fld[0] and @fld = @{$fld[0]};
    (my $xff = join "\xff " => @fld) =~ s/\xff( (a|de)sc(\xff|$))/\U$1/gi;
    @fld = split m/\xff/ => $xff;
    my $nm = "I_${tbl}_" . join ("", map { (my $x = $_) =~ s/\s.*//; $x } @fld);
    length ($nm) > 30 and substr ($nm, 30) = "";
    if ($ENV{DBTYPE} eq "ORACLE") {
	my $ts = (split m:/:, $ENV{DBUSER}||$ENV{ORACLE_USERID})[0];
	local $" = ", ";
	$idx = join "\n",
	    "create$unique index $nm",
	    "  on $tbl (@fld)",
	    "  tablespace $ts",
	    "  storage (pctincrease 0)";
	}
    else {
	local $" = ", ";
	$idx = join "\n",
	    "create$unique btree index $nm",
	    "  on $tbl (@fld)";
	}
    $opt_v and print STDERR $idx, "\n";
    $dbh->do ($idx);
    } # create_index

1;
