package PROCURA::Diac;

$PROCURA::Diac::VERSION = 4.15;

=head1 NAME

Diac - supply perl-Diac interface

=head1 SYNOPSIS

 use PROCURA::Diac;

 $extraspace = gbadiaccnt ("Francois");      # result: 1
 $reallength = diac_length ("Francois");     # result: 8 (length = 9)

 $e010240 = diac_substr ("Francois", 1);     # result: "rancois"
 $e010240 = diac_substr ("Francois", 2, 4);  # result: "anco"
 $e010240 = diac_substr ("Francois", 3, 8);  # result: "ncois   "

 $e010240 = diacgba ("Francois", "\x05\x5B"); # result: "Francois"
 @warnings = diacgba_check ("Francois", "\x06\x5B\x3F\x0A\x06\x06\xd2\x07")
    => ("Base char 'o' for diac  91 on pos  6 should be 'c'",
        "Diac  63 on pos 10 past stringlength (8)",
        "Diac   6 on pos  6 overloads  91 on same pos",
        "Diac 210 on pos  7 unsupported",
        );

 ($db, $d) = diacdbd ("Francois");           # reverse of diacgba
 ($db, $d) = diacdbd ("Frncois", 0);        #  => (undef, "diag message")
 ($db, $d) = diacdbd ("Frncois", 1);        #  => ("Frncois", "\x05\x5B")
 ($db, $d) = diacdbd ("Frncois", "1!");     #  same + warning
 ($db, $d) = diacdbd ("Frncois", "?");      #  => ("Fr?ncois", "\x05\x5B")
 ($db, $d) = diacdbd ("Frncois", "?!");     #  same + warning
 ($db, $d) = diacdbx ("Francois");           #  => ("Francois", "5:91")
 ($cs7, undef) = diacdbd ("Francois");       # Strip all diacs

 $cs7 = TexNone    ("Francois");             #  same, result: "Francois"
 $cs8 = Texcs8     ("Francois");             # Convert to multinational cs8
 $ocr = TexOCRB    ("Francois");             # Convert to OCRB
 $iso = Tex8859_1  ("Francois");             # Convert to ISO8859_1, LATIN1
 $iso = Tex8859_9  ("Francois");             # Convert to ISO8859_9, LATIN5
 $uni = TexUnicode ("Francois");             # Convert to Unicode

 $tex = UnicodeTex ("Franois");              # Convert Unicode to TELETEX

 $e010240 = "Fran\xA7ois";
 CnvDiacCP500 (\$e010240);                    # result: "Francois"

 $e010240 = "Fran\xE7ois";
 CnvDiac8859_1 (\$e010240);                   # result: "Francois"
 CnvDiac8859_9 (\$e010240);                   # result: "Francois"

 my $desc = DiacLookup ("utf8", "")->[2];    # result: c_CEDILLA
 $r = DiacLookup ("utf8", "c_CEDILLA");       # [ "c", 91, "c_CEDILLA",
                                              #   "\x{e7}" ]

=head1 DESCRIPTION

=over 4

=item

=item B<gbadiaccnt (>C<$teletex_var>B<)>

Count number of characters gained when variable is displayed or
printed.

=item B<diac_length (>C<$teletex_var>B<)>

Return the length of the string as if all diacs would have been
resolved.

=item B<diac_substr (>C<$teletex_var, $start [, $length]>B<)>

As substr would, but taking diacs into account.

If $length is greater than the resulting teletex string, the result
is padded with space to the right to meet the requested length.

=item B<diacgba (>C<$db_var, $db_diacs>B<)>

Convert a set of base characters and diac's to teletex. Both $db_var
and $db_diacs are typically directly extracted from the database.

=item (C<$db_var>, C<$db_diacs>) = B<diacdbd (>C<$var>[, C<$force>]B<)>

=item (C<$db_var>, C<$db_diacs>) = B<diacdbx (>C<$var>[, C<$force>]B<)>

Convert a teletex string to a set of base characters and diac's. Both
$db_var and $db_diacs are typically directly inserted into the database.

If an error occurs during conversion, C<$db_var> is undefined and
C<$db_diacs> contains a diagnostic message. If C<$force> is 1 (default 0),
the invalid sequence is retained. If C<$force> is true but not 1, all
invalid characters are replaced with the sequence in C<$force>. If C<$force>
is trailed with an exclamation mark C<!>, it is removed from C<$force>,
and warnings will be issued on illegal conversions (to STDERR).

When using B<diacdbx ()>, C<$d> is returned in portable ascii format
as comma separated pairs of position:diac_num where position is one-based
decimal and diac_num is decimal.

=item C<$tex> = B<CnvDiacCP500 (>C<\$var>B<);>

Convert Code Page 500 EBCDIC->ASCII-translation to TELETEX.
$var has already been converted from EBCDIC to ASCII (i.e using
"dd conv=ascii").

=item C<$tex> = B<CnvDiac8859_1 (>C<\$var>B<);>

Convert ISO 8859-1 (LATIN1) character set to TELETEX.

=item C<$tex> = B<CnvDiac8859_9 (>C<\$var>B<);>

Convert ISO 8859-9 (LATIN5) character set to TELETEX.

=item C<$cs7> = B<TexNone (>C<\$var>B<);>

Strip all diacritical marks from TELETEX to plain ASCII.

=item C<$cs8> = B<Texcs8 (>C<\$var>B<);>

Convert TELETEX to multinational cs8 (Wyse60).

=item C<$ocr> = B<TexOCRB (>C<\$var>B<);>

Convert TELETEX to OCRB.

=item C<$iso> = B<Tex8859_1 (>C<\$var>B<);>

Convert TELETEX to ISO 8859-1 (Latin 1).

=item C<$iso> = B<Tex8859_9 (>C<\$var>B<);>

Convert TELETEX to ISO 8859-9 (Latin 5).

=item C<$uni> = B<TexUnicode (>C<\$var>B<);>

Convert TELETEX to Unicode.

=item C<$tex> = B<UnicodeTex (>C<\$var>B<);>

Convert Unicode to TELETEX.

=back

=cut

### ###########################################################################

use Exporter;
use vars qw(@ISA @EXPORT);
use charnames ":alias" => ":pro";

@ISA	 = qw(Exporter);
@EXPORT	 = qw(diac_debug
              gbadiaccnt diac_length
	      diac_substr
	      diacgba diacdbd diacdbx  diacgba_check
	      CnvDiacCP500 CnvDiac8859_1 CnvDiac8859_9
	      TexNone Texcs8 TexOCRB
	      Tex8859_1 Tex8859_9 TexUnicode UnicodeTex
	      DiacLookup
	      );

use strict;

use vars qw($debug);
$debug = 0;

my @dsw = (
#   0 1 2 3 4 5 6 7 8 9 A B C D E F
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	# 0
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	# 1
    1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,	# 2
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,	# 3
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,	# 4
    1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,	# 5
    0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,	# 6
    1,1,1,1,1,1,1,1,1,1,1,0,1,0,0,0,	# 7
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	# 8
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	# 9
    0,1,1,1,1,1,1,1,1,0,0,1,0,0,0,0,	# A
    1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,	# B
    0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,	# C
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	# D
    1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,	# E
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0);	# F

my %dbd_tab;
my %uni_tab;
my %gba_tab;
my @gba_tab = (
    [ 1, '?', undef,	"A_MAGNIET"		],	# A
    [ 2, 'A', chr 0xC1,	"A_GRAVE"		],
    [ 2, 'A', chr 0xC3,	"A_CIRCUM"		],
    [ 2, 'A', chr 0xC8,	"A_DIAERES"		],
    [ 2, 'A', chr 0xC4,	"A_TILDE"		],
    [ 2, 'A', chr 0xC6,	"A_BREVE"		],
    [ 2, 'A', chr 0xCA,	"A_RING"		],
    [ 2, 'A', chr 0xC5,	"A_MACRON"		],
    [ 2, 'A', chr 0xCE,	"A_OGONEK"		],
    [ 2, 'C', chr 0xC2,	"C_ACUTE"		],	# C
    [ 2, 'C', chr 0xC3,	"C_CIRCUM"		],
    [ 2, 'C', chr 0xCF,	"C_CARON"		],
    [ 2, 'C', chr 0xC7,	"C_DOT"			],
    [ 2, 'C', chr 0xCB,	"C_CEDILLA"		],
    [ 2, 'D', chr 0xCF,	"D_CARON"		],	# D
    [ 2, 'E', chr 0xC2,	"E_ACUTE"		],	# E
    [ 2, 'E', chr 0xC1,	"E_GRAVE"		],
    [ 2, 'E', chr 0xC3,	"E_CIRCUM"		],
    [ 2, 'E', chr 0xC8,	"E_DIAERES"		],
    [ 2, 'E', chr 0xCF,	"E_CARON"		],
    [ 2, 'E', chr 0xC7,	"E_DOT"			],
    [ 2, 'E', chr 0xC5,	"E_MACRON"		],
    [ 2, 'E', chr 0xCE,	"E_OGONEK"		],
    [ 2, 'G', chr 0xC3,	"G_CIRCUM"		],	# G
    [ 2, 'G', chr 0xC6,	"G_BREVE"		],
    [ 2, 'G', chr 0xC7,	"G_DOT"			],
    [ 2, 'G', chr 0xCB,	"G_CEDILLA"		],
    [ 2, 'H', chr 0xC3,	"H_CIRCUM"		],	# H
    [ 2, 'I', chr 0xC2,	"I_ACUTE"		],	# I
    [ 2, 'I', chr 0xC1,	"I_GRAVE"		],
    [ 2, 'I', chr 0xC3,	"I_CIRCUM"		],
    [ 2, 'I', chr 0xC8,	"I_DIAERES"		],
    [ 2, 'I', chr 0xC4,	"I_TILDE"		],
    [ 2, 'I', chr 0xC7,	"I_DOT"			],
    [ 2, 'I', chr 0xC5,	"I_MACRON"		],
    [ 2, 'I', chr 0xCE,	"I_OGONEK"		],
    [ 2, 'J', chr 0xC3,	"J_CIRCUM"		],	# J
    [ 2, 'K', chr 0xCB,	"K_CEDILLA"		],	# K
    [ 2, 'L', chr 0xC2,	"L_ACUTE"		],	# L
    [ 2, 'L', chr 0xCF,	"L_CARON"		],
    [ 2, 'L', chr 0xCB,	"L_CEDILLA"		],
    [ 2, 'N', chr 0xC2,	"N_ACUTE"		],	# N
    [ 2, 'N', chr 0xC4,	"N_TILDE"		],
    [ 2, 'N', chr 0xCF,	"N_CARON"		],
    [ 2, 'N', chr 0xCB,	"N_CEDILLA"		],
    [ 2, 'O', chr 0xC2,	"O_ACUTE"		],	# O
    [ 2, 'O', chr 0xC1,	"O_GRAVE"		],
    [ 2, 'O', chr 0xC3,	"O_CIRCUM"		],
    [ 2, 'O', chr 0xC8,	"O_DIAERES"		],
    [ 2, 'O', chr 0xC4,	"O_TILDE"		],
    [ 2, 'O', chr 0xCD,	"O_DBLACUTE"		],
    [ 2, 'O', chr 0xC5,	"O_MACRON"		],
    [ 2, 'R', chr 0xC2,	"R_ACUTE"		],	# R
    [ 2, 'R', chr 0xCF,	"R_CARON"		],
    [ 2, 'R', chr 0xCB,	"R_CEDILLA"		],
    [ 2, 'S', chr 0xC2,	"S_ACUTE"		],	# S
    [ 2, 'S', chr 0xC3,	"S_CIRCUM"		],
    [ 2, 'S', chr 0xCF,	"S_CARON"		],
    [ 2, 'S', chr 0xCB,	"S_CEDILLA"		],
    [ 2, 'T', chr 0xCF,	"T_CARON"		],	# T
    [ 2, 'T', chr 0xCB,	"T_CEDILLA"		],
    [ 2, 'U', chr 0xC2,	"U_ACUTE"		],	# U
    [ 2, 'U', chr 0xC1,	"U_GRAVE"		],
    [ 2, 'U', chr 0xC3,	"U_CIRCUM"		],
    [ 2, 'U', chr 0xC8,	"U_DIAERES"		],
    [ 2, 'U', chr 0xC4,	"U_TILDE"		],
    [ 2, 'U', chr 0xC6,	"U_BREVE"		],
    [ 2, 'U', chr 0xCD,	"U_DBLACUTE"		],
    [ 2, 'U', chr 0xCA,	"U_RING"		],
    [ 2, 'U', chr 0xC5,	"U_MACRON"		],
    [ 2, 'U', chr 0xCE,	"U_OGONEK"		],
    [ 2, 'W', chr 0xC3,	"W_CIRCUM"		],	# W
    [ 2, 'Y', chr 0xC2,	"Y_ACUTE"		],	# Y
    [ 2, 'Y', chr 0xC3,	"Y_CIRCUM"		],
    [ 2, 'Y', chr 0xC8,	"Y_DIAERES"		],
    [ 2, 'Z', chr 0xC2,	"Z_ACUTE"		],	# Z
    [ 2, 'Z', chr 0xCF,	"Z_CARON"		],
    [ 2, 'Z', chr 0xC7,	"Z_DOT"			],
    [ 2, 'a', chr 0xC2,	"a_ACUTE"		],	# a
    [ 2, 'a', chr 0xC1,	"a_GRAVE"		],
    [ 2, 'a', chr 0xC3,	"a_CIRCUM"		],
    [ 2, 'a', chr 0xC8,	"a_DIAERES"		],
    [ 2, 'a', chr 0xC4,	"a_TILDE"		],
    [ 2, 'a', chr 0xC6,	"a_BREVE"		],
    [ 2, 'a', chr 0xCA,	"a_RING"		],
    [ 2, 'a', chr 0xC5,	"a_MACRON"		],
    [ 2, 'a', chr 0xCE,	"a_OGONEK"		],
    [ 2, 'c', chr 0xC2,	"c_ACUTE"		],	# c
    [ 2, 'c', chr 0xC3,	"c_CIRCUM"		],
    [ 2, 'c', chr 0xCF,	"c_CARON"		],
    [ 2, 'c', chr 0xC7,	"c_DOT"			],
    [ 2, 'c', chr 0xCB,	"c_CEDILLA"		],
    [ 2, 'd', chr 0xCF,	"d_CARON"		],	# d
    [ 2, 'e', chr 0xC2,	"e_ACUTE"		],	# e
    [ 2, 'e', chr 0xC1,	"e_GRAVE"		],
    [ 2, 'e', chr 0xC3,	"e_CIRCUM"		],
    [ 2, 'e', chr 0xC8,	"e_DIAERES"		],
    [ 2, 'e', chr 0xCF,	"e_CARON"		],
    [ 2, 'e', chr 0xC7,	"e_DOT"			],
    [ 2, 'e', chr 0xC5,	"e_MACRON"		],
    [ 2, 'e', chr 0xCE,	"e_OGONEK"		],
    [ 2, 'g', chr 0xC2,	"g_ACUTE"		],	# g
    [ 2, 'g', chr 0xC3,	"g_CIRCUM"		],
    [ 2, 'g', chr 0xC6,	"g_BREVE"		],
    [ 2, 'g', chr 0xC7,	"g_DOT"			],
    [ 2, 'h', chr 0xC3,	"h_CIRCUM"		],	# h
    [ 2, 'i', chr 0xC2,	"i_ACUTE"		],	# i
    [ 2, 'i', chr 0xC1,	"i_GRAVE"		],
    [ 2, 'i', chr 0xC3,	"i_CIRCUM"		],
    [ 2, 'i', chr 0xC8,	"i_DIAERES"		],
    [ 2, 'i', chr 0xC4,	"i_TILDE"		],
    [ 2, 'i', chr 0xC5,	"i_MACRON"		],
    [ 2, 'i', chr 0xCE,	"i_OGONEK"		],
    [ 2, 'j', chr 0xC3,	"j_CIRCUM"		],	# j
    [ 2, 'k', chr 0xCB,	"k_CEDILLA"		],	# k
    [ 2, 'l', chr 0xC2,	"l_ACUTE"		],	# l
    [ 2, 'l', chr 0xCF,	"l_CARON"		],
    [ 2, 'l', chr 0xCB,	"l_CEDILLA"		],
    [ 2, 'n', chr 0xC2,	"n_ACUTE"		],	# n
    [ 2, 'n', chr 0xC4,	"n_TILDE"		],
    [ 2, 'n', chr 0xCF,	"n_CARON"		],
    [ 2, 'n', chr 0xCB,	"n_CEDILLA"		],
    [ 2, 'o', chr 0xC2,	"o_ACUTE"		],	# o
    [ 2, 'o', chr 0xC1,	"o_GRAVE"		],
    [ 1, '?', undef,	"o_MAG_NIET_SQL"	],
    [ 2, 'o', chr 0xC3,	"o_CIRCUM"		],
    [ 2, 'o', chr 0xC8,	"o_DIAERES"		],
    [ 2, 'o', chr 0xC4,	"o_TILDE"		],
    [ 2, 'o', chr 0xCD,	"o_DBLACUTE"		],
    [ 2, 'o', chr 0xC5,	"o_MACRON"		],
    [ 2, 'r', chr 0xC2,	"r_ACUTE"		],	# r
    [ 2, 'r', chr 0xCF,	"r_CARON"		],
    [ 2, 'r', chr 0xCB,	"r_CEDILLA"		],
    [ 2, 's', chr 0xC2,	"s_ACUTE"		],	# s
    [ 2, 's', chr 0xC3,	"s_CIRCUM"		],
    [ 2, 's', chr 0xCF,	"s_CARON"		],
    [ 2, 's', chr 0xCB,	"s_CEDILLA"		],
    [ 2, 't', chr 0xCF,	"t_CARON"		],	# t
    [ 2, 't', chr 0xCB,	"t_CEDILLA"		],
    [ 2, 'u', chr 0xC2,	"u_ACUTE"		],	# u
    [ 2, 'u', chr 0xC1,	"u_GRAVE"		],
    [ 2, 'u', chr 0xC3,	"u_CIRCUM"		],
    [ 2, 'u', chr 0xC8,	"u_DIAERES"		],
    [ 2, 'u', chr 0xC4,	"u_TILDE"		],
    [ 2, 'u', chr 0xC6,	"u_BREVE"		],
    [ 2, 'u', chr 0xCD,	"u_DBLACUTE"		],
    [ 2, 'u', chr 0xCA,	"u_RING"		],
    [ 2, 'u', chr 0xC5,	"u_MACRON"		],
    [ 2, 'u', chr 0xCE,	"u_OGONEK"		],
    [ 2, 'w', chr 0xC3,	"w_CIRCUM"		],	# w
    [ 2, 'y', chr 0xC2,	"y_ACUTE"		],	# y
    [ 2, 'y', chr 0xC3,	"y_CIRCUM"		],
    [ 2, 'y', chr 0xC8,	"y_DIAERES"		],
    [ 2, 'z', chr 0xC2,	"z_ACUTE"		],	# z
    [ 2, 'z', chr 0xCF,	"z_CARON"		],
    [ 2, 'z', chr 0xC7,	"z_DOT"			],
    [ 2, 'A', chr 0xC2,	"A_ACUTE"		],

    [ 1, 'c', chr 0xA2,	"CENT_IDX"		],
    [ 1, 'P', chr 0xA3,	"POUND_IDX"		],
    [ 1, '$', chr 0xA4,	"DOLLAR_IDX"		],
    [ 1, 'Y', chr 0xA5,	"YEN_IDX"		],
    [ 1, '#', chr 0xA6,	"NUMBER_IDX"		],
    [ 1, '$', chr 0xA7,	"SECTION_IDX"		],
    [ 1, '@', chr 0xA8,	"CURRENCY_IDX"		],
    [ 1, '<', chr 0xAB,	"AQML_IDX"		],
    [ 1, 'o', chr 0xB0,	"DEGREE_IDX"		],
    [ 1, '~', chr 0xB1,	"PL_MIN_IDX"		],
    [ 1, '2', chr 0xB2,	"SUP2_IDX"		],
    [ 1, '3', chr 0xB3,	"SUP3_IDX"		],
    [ 1, 'x', chr 0xB4,	"MULTIPLY_IDX"		],
    [ 1, 'm', chr 0xB5,	"MICRO_IDX"		],
    [ 1, 'P', chr 0xB6,	"PILCROW_IDX"		],
    [ 1, '.', chr 0xB7,	"MDOT_IDX"		],
    [ 1, '/', chr 0xB8,	"DIVIDE_IDX"		],
    [ 1, '>', chr 0xBB,	"AQMR_IDX"		],
    [ 1, '%', chr 0xBC,	"QUARTER1_IDX"		],
    [ 1, '%', chr 0xBD,	"HALF_IDX"		],
    [ 1, '%', chr 0xBE,	"QUARTER3_IDX"		],
    [ 1, '?', chr 0xBF,	"IQML_IDX"		],
    [ 1, 'O', chr 0xE0,	"OHM_IDX"		],
    [ 1, 'A', chr 0xE1,	"CAP_AE_IDX"		],
    [ 1, 'D', chr 0xE2,	"DSTROKE_IDX"		],
    [ 1, 'a', chr 0xE3,	"ORDIFEM_IDX"		],
    [ 1, 'H', chr 0xE4,	"HSTROKE_IDX"		],
    [ 1, 'Y', chr 0xE6,	"CAP_IJ_IDX"		],
    [ 1, 'L', chr 0xE7,	"LMDOT_IDX"		],
    [ 1, 'L', chr 0xE8,	"LSTROKE_IDX"		],
    [ 1, 'O', chr 0xE9,	"OSLASH_IDX"		],
    [ 1, 'O', chr 0xEA,	"CAP_OE_IDX"		],
    [ 1, 'o', chr 0xEB,	"ORDIMAS_IDX"		],
    [ 1, 'P', chr 0xEC,	"CAP_THORN_IDX"		],
    [ 1, 'T', chr 0xED,	"TSTROKE_IDX"		],
    [ 1, 'N', chr 0xEE,	"CAP_ENG_IDX"		],
    [ 1, 'n', chr 0xEF,	"NAPOS_IDX"		],
    [ 1, 'k', chr 0xF0,	"SMALL_K_IDX"		],
    [ 1, 'a', chr 0xF1,	"SMALL_AE_IDX"		],
    [ 1, 'd', chr 0xF2,	"dSTROKE_IDX"		],
    [ 1, 'd', chr 0xF3,	"ETH_IDX"		],
    [ 1, 'h', chr 0xF4,	"hSTROKE_IDX"		],
    [ 1, 'i', chr 0xF5,	"INODOT_IDX"		],
    [ 1, 'y', chr 0xF6,	"SMALL_IJ_IDX"		],
    [ 1, 'l', chr 0xF7,	"lMDOT_IDX"		],
    [ 1, 'l', chr 0xF8,	"lSTROKE_IDX"		],
    [ 1, 'o', chr 0xF9,	"oSLASH_IDX"		],
    [ 1, 'o', chr 0xFA,	"SMALL_OE_IDX"		],
    [ 1, 'B', chr 0xFB,	"RINGELS_IDX"		],
    [ 1, 'p', chr 0xFC,	"SMALL_THORN_IDX"	],
    [ 1, 't', chr 0xFD,	"tSTROKE_IDX"		],
    [ 1, 'n', chr 0xFE,	"SMALL_ENG_IDX"		],
    [ 1, '!', chr 0xA1,	"IEM_IDX"		],
    [ 1, '?', "_\b",	"NSU"			]	# entry voor DIAC_UNKNOWN
    );

my %DiacCP500 = (
    "\x9F" => "\xB8",		# divide sign B8
    "\xA0" => "\xFF",		#
    "\xA1" => "\xC3\x61",	# a circumflex C361
    "\xA2" => "\xC8\x61",	# a diaeresis C861
    "\xA3" => "\xC1\x61",	# a grave C161
    "\xA4" => "\xC2\x61",	# a acute C261
    "\xA5" => "\xC4\x61",	# a tilde C461
    "\xA6" => "\xCA\x61",	# a ring CA61
    "\xA7" => "\xCB\x63",	# c cedilla CB63
    "\xA8" => "\xC4\x6E",	# n tilde C46E
    "\xA9" => "\xC2\x65",	# e acute C265
    "\xAA" => "\xC3\x65",	# e circumflex C365
    "\xAB" => "\xC8\x65",	# e diaeresis C865
    "\xAC" => "\xC1\x65",	# e grave C165
    "\xAD" => "\xC2\x69",	# i acute C269
    "\xAE" => "\xC3\x69",	# i circumflex C369
    "\xAF" => "\xC8\x69",	# i diaeresis C869
    "\xB0" => "\xC1\x69",	# i grave C169
    "\xB1" => "\xFB",		# sharp s, German FB
    "\xB2" => "\xC3\x41",	# A circumflex C341
    "\xB3" => "\xC8\x41",	# A diaeresis C841
    "\xB4" => "\xC1\x41",	# A grave C141
    "\xB5" => "\xC2\x41",	# A acute C241
    "\xB6" => "\xC4\x41",	# A tilde C441
    "\xB7" => "\xCA\x41",	# A ring CA41
    "\xB8" => "\xCB\x43",	# C cedilla CB43
    "\xB9" => "\xC4\x4E",	# N tilde C44E
    "\xBA" => "\xF9",		# o with slash F9
    "\xBB" => "\xC2\x45",	# E acute C245
    "\xBC" => "\xC3\x45",	# E circumflex C345
    "\xBD" => "\xC8\x45",	# E diaeresis C845
    "\xBE" => "\xC1\x45",	# E grave C145
    "\xBF" => "\xC2\x49",	# I acute C249
    "\xC0" => "\xC3\x49",	# I circumflex C349
    "\xC1" => "\xC8\x49",	# I diaeresis C849
    "\xC2" => "\xC1\x49",	# I grave C149
    "\xC3" => "\xE9",		# O with slash E9
    "\xC4" => "\xAB",		# angle quotation mark left AB
    "\xC5" => "\xBB",		# angle quotation mark left BB
    "\xC6" => "\xF3",		# small eth, Icelandic F3
    "\xC7" => "\xC2\x79",	# y acute C279
    "\xC8" => "\xFC",		# small thorn, Icelandic FC
    "\xC9" => "\xB1",		# plus/minus sign B1
    "\xCA" => "\xB0",		# degree sign B0
    "\xCB" => "\xE3",		# ordinal indicator, feminine E3
    "\xCC" => "\xEB",		# ordinal indicator, masculine EB
    "\xCD" => "\xF1",		# ae diphthong F1
    "\xCF" => "\xE1",		# AE diphthong E1
    "\xD0" => "\xA8",		# currency symbol A8
    "\xD1" => "\xB5",		# micro sign B5
    "\xD2" => "\xA1",		# inverted exclamation mark A1
    "\xD3" => "\xBF",		# inverted question mark, left BF
    "\xD4" => "\xE2",		# d with stroke E2
    "\xD5" => "\xC2\x79",	# Y acute C259
    "\xD6" => "\xEC",		# capital thorn, Icelandic EC
    "\xD7" => "\xD2",		# register sign D2
    "\xD8" => "\xA2",		# cent sign A2
    "\xD9" => "\xA3",		# pound sign A3
    "\xDA" => "\xA5",		# yen sign A5
    "\xDB" => "\xB7",		# middle dot B7
    "\xDC" => "\xD3",		# copyright sign D3
    "\xDD" => "\xA7",		# section sign A7
    "\xDE" => "\xB6",		# paragraph sign, pilcrow B6
    "\xDF" => "\xBC",		# fraction one quarter BC
    "\xE0" => "\xBD",		# fraction one half BD
    "\xE1" => "\xBE",		# fraction three quarters BE
    "\xE5" => "\xC9",		# double dot C9
    "\xE6" => "\xB9",		# quote B9
    "\xE7" => "\xB4",		# multiply sign B4
    "\xE8" => "\xCC",		# low line CC
    "\xE9" => "\xC3\x6F",	# o circumflex C36F
    "\xEA" => "\xC8\x6F",	# o diaeresis C86F
    "\xEB" => "\xC1\x6F",	# o grave C16F
    "\xEC" => "\xC2\x6F",	# o acute C26F
    "\xED" => "\xC4\x6F",	# o tilde C46F
    "\xEE" => "\xD1",		# superscript 1 D1
    "\xEF" => "\xC3\x75",	# u circumflex C375
    "\xF0" => "\xC8\x75",	# u diaeresis C875
    "\xF1" => "\xC1\x75",	# u grave C175
    "\xF2" => "\xC2\x75",	# u acute C275
    "\xF3" => "\xC8\x79",	# y diaeresis C879
    "\xF4" => "\xB2",		# superscript 2 B2
    "\xF5" => "\xC3\x4F",	# O circumflex C34F
    "\xF6" => "\xC8\x4F",	# O diaeresis C84F
    "\xF7" => "\xC1\x4F",	# O grave C14F
    "\xF8" => "\xC2\x4F",	# O acute C24F
    "\xF9" => "\xC4\x4F",	# O tilde C44F
    "\xFA" => "\xB3",		# superscript 3 B3
    "\xFB" => "\xC3\x55",	# U circumflex C355
    "\xFC" => "\xC8\x55",	# U diaeresis C855
    "\xFD" => "\xC1\x55",	# U grave C155
    "\xFE" => "\xC2\x55"	# U acute C255
    );

my %Diac8859_9 = (
    "\xA0" => "\x20",		#
    "\xA1" => "\xA1",		#  IEM_IDX
    "\xA2" => "\xA2",		#  CENT_IDX
    "\xA3" => "\xA3",		#  POUND_IDX
    "\xA4" => "\xA8",		#  CURRENCY_IDX
    "\xA5" => "\xA5",		#  YEN_IDX
    "\xA6" => "\x7C",		# |
    "\xA7" => "\xA7",		#  SECTION_IDX
    "\xA8" => "\x20",		#
    "\xA9" => "\xD3",		#   copyright sign D3
    "\xAA" => "\xE3",		#  ORDIFEM_IDX
    "\xAB" => "\xAB",		#  AQML_IDX
    "\xAC" => "\x20",		#
    "\xAD" => "\x20",		#
    "\xAE" => "\x20",		#
    "\xAF" => "\x20",		#
    "\xB0" => "\xB0",		#  DEGREE_IDX
    "\xB1" => "\xB1",		#  PL_MIN_IDX
    "\xB2" => "\xB2",		#  SUP2_IDX
    "\xB3" => "\xB3",		#  SUP3_IDX
    "\xB4" => "\x20",		#
    "\xB5" => "\xB5",		#  MICRO_IDX
    "\xB6" => "\xB6",		#  PILCROW_IDX
    "\xB7" => "\xB7",		#  MDOT_IDX
    "\xB8" => "\x20",		#
    "\xB9" => "\xD1",		#   superscript 1 D1
    "\xBA" => "\xEB",		#  ORDIMAS_IDX
    "\xBB" => "\xBB",		#  AQMR_IDX
    "\xBC" => "\xBC",		#  QUARTER1_IDX
    "\xBD" => "\xBD",		#  HALF_IDX
    "\xBE" => "\xBE",		#  QUARTER3_IDX
    "\xBF" => "\xBF",		#  IQML_IDX
    "\xC0" => "\xC1\x41",	#  A_GRAVE
    "\xC1" => "\xC2\x41",	#  A_ACUTE
    "\xC2" => "\xC3\x41",	#  A_CIRCUM
    "\xC3" => "\xC4\x41",	#  A_TILDE
    "\xC4" => "\xC8\x41",	#  A_DIAERES
    "\xC5" => "\xCA\x41",	#  A_RING
    "\xC6" => "\xE1",		#  CAP_AE_IDX
    "\xC7" => "\xCB\x43",	#  C_CEDILLA
    "\xC8" => "\xC1\x45",	#  E_GRAVE
    "\xC9" => "\xC2\x45",	#  E_ACUTE
    "\xCA" => "\xC3\x45",	#  E_CIRCUM
    "\xCB" => "\xC8\x45",	#  E_DIAERES
    "\xCC" => "\xC1\x49",	#  I_GRAVE
    "\xCD" => "\xC2\x49",	#  I_ACUTE
    "\xCE" => "\xC3\x49",	#  I_CIRCUM
    "\xCF" => "\xC8\x49",	#  I_DIAERES
    "\xD0" => "\xC6\x47",	#  G_BREVE
    "\xD1" => "\xC4\x4E",	#  N_TILDE
    "\xD2" => "\xC1\x4F",	#  O_GRAVE
    "\xD3" => "\xC2\x4F",	#  O_ACUTE
    "\xD4" => "\xC3\x4F",	#  O_CIRCUM
    "\xD5" => "\xC4\x4F",	#  O_TILDE
    "\xD6" => "\xC8\x4F",	#  O_DIAERES
    "\xD7" => "\xB4",		#  MULTIPLY_IDX
    "\xD8" => "\xE9",		#  OSLASH_IDX
    "\xD9" => "\xC1\x55",	#  U_GRAVE
    "\xDA" => "\xC2\x55",	#  U_ACUTE
    "\xDB" => "\xC3\x55",	#  U_CIRCUM
    "\xDC" => "\xC8\x55",	#  U_DIAERES
    "\xDD" => "\xC7\x49",	#  I_DOT
    "\xDE" => "\xCB\x53",	#  S_CEDILLA
    "\xDF" => "\xFB",		#  RINGELS_IDX
    "\xE0" => "\xC1\x61",	#  a_GRAVE
    "\xE1" => "\xC2\x61",	#  a_ACUTE
    "\xE2" => "\xC3\x61",	#  a_CIRCUM
    "\xE3" => "\xC4\x61",	#  a_TILDE
    "\xE4" => "\xC8\x61",	#  a_DIAERES
    "\xE5" => "\xCA\x61",	#  a_RING
    "\xE6" => "\xF1",		#  SMALL_AE_IDX
    "\xE7" => "\xCB\x63",	#  c_CEDILLA
    "\xE8" => "\xC1\x65",	#  e_GRAVE
    "\xE9" => "\xC2\x65",	#  e_ACUTE
    "\xEA" => "\xC3\x65",	#  e_CIRCUM
    "\xEB" => "\xC8\x65",	#  e_DIAERES
    "\xEC" => "\xC1\x69",	#  i_GRAVE
    "\xED" => "\xC2\x69",	#  i_ACUTE
    "\xEE" => "\xC3\x69",	#  i_CIRCUM
    "\xEF" => "\xC8\x69",	#  i_DIAERES
    "\xF0" => "\xC6\x67",	#  g_BREVE
    "\xF1" => "\xC4\x6E",	#  n_TILDE
    "\xF2" => "\xC1\x6F",	#  o_GRAVE
    "\xF3" => "\xC2\x6F",	#  o_ACUTE
    "\xF4" => "\xC3\x6F",	#  o_CIRCUM
    "\xF5" => "\xC4\x6F",	#  o_TILDE
    "\xF6" => "\xC8\x6F",	#  o_DIAERES
    "\xF7" => "\xB8",		#  DIVIDE_IDX
    "\xF8" => "\xF9",		#  oSLASH_IDX
    "\xF9" => "\xC1\x75",	#  u_GRAVE
    "\xFA" => "\xC2\x75",	#  u_ACUTE
    "\xFB" => "\xC3\x75",	#  u_CIRCUM
    "\xFC" => "\xC8\x75",	#  u_DIAERES
    "\xFD" => "\xF5",		#  INODOT_IDX
    "\xFE" => "\xCB\x73",	#  s_CEDILLA
    "\xFF" => "\xC8\x79",	#  y_DIAERES
    );
my %Diac8859_1 = %Diac8859_9;
@Diac8859_1{"\xD0", "\xDD", "\xDE", "\xF0", "\xFD", "\xFE"} = (
    "\xE2",	#  DSTROKE_IDX
    "\xC2\x79",	#  Y_ACUTE
    "\xEC",	#  CAP_THORN_IDX
    "\xF3",	#  ETH_IDX
    "\xC2\x59",	#  y_ACUTE
    "\xFC",	#  SMALL_THORN_IDX
    );

my %cs = (
    cs7       => [	# None
	(undef) x $#gba_tab
	],
    cs8       => [	# Standard multinational cs8 (wy60)
	undef,  "A",    "A",    "\216", "A",    "A",    "\217", "A",    "A",
	"C",    "C",    "C",    "C",    "C",    "D",    "\220", "E",    "E",
	"E",    "E",    "E",    "E",    "E",    "G",    "G",    "G",    "G",
	"H",    "I",    "I",    "I",    "I",    "I",    "I",    "I",    "I",
	"J",    "K",    "L",    "L",    "L",    "N",    "\245", "N",    "N",
	"O",    "O",    "O",    "\231", "O",    "O",    "O",    "R",    "R",
	"R",    "S",    "S",    "S",    "S",    "T",    "T",    "U",    "U",
	"U",    "\232", "U",    "U",    "U",    "U",    "U",    "U",    "W",
	"Y",    "Y",    "Y",    "Z",    "Z",    "Z",    "\240", "\205", "\203",
	"\204", "a",    "a",    "\206", "a",    "a",    "c",    "c",    "c",
	"c",    "\207", "d",    "\202", "\212", "\210", "\211", "e",    "e",
	"e",    "e",    "g",    "g",    "g",    "g",    "h",    "\241", "\215",
	"\214", "\213", "i",    "i",    "i",    "j",    "k",    "l",    "l",
	"l",    "n",    "\244", "n",    "n",    "\242", "\225", undef,  "\223",
	"\224", "o",    "o",    "o",    "r",    "r",    "r",    "s",    "s",
	"s",    "s",    "t",    "t",    "\243", "\227", "\226", "\201", "u",
	"u",    "u",    "u",    "u",    "u",    "w",    "y",    "y",    "\230",
	"z",    "z",    "z",    "A",    "\233", "\234", undef,  "\235", undef,
	"S",    undef,  "\256", "\370", undef,  "\375", undef,  undef,  "\346",
	undef,  "\372", "\366", "\257", "\254", "\253", undef,  "\250", "\352",
	"\222", undef,  "\246", undef,  undef,  undef,  undef,  undef,  undef,
	"\247", undef,  undef,  undef,  undef,  undef,  "\221", undef,  undef,
	undef,  undef,  undef,  undef,  undef,  undef,  undef,  undef,  undef,
	undef,  "\374", "\255", undef,  undef
	],
    iso8859_1 => [	# ISO8859_1, LATIN1
	undef,  "\300", "\302", "\304", "\303", "A",    "\305", "A",    "A",
	"C",    "C",    "C",    "C",    "\307", "D",    "\311", "\310", "\312",
	"\313", "E",    "E",    "E",    "E",    "G",    "G",    "G",    "G",
	"H",    "\315", "\314", "\316", "\317", "I",    "I",    "I",    "I",
	"J",    "K",    "L",    "L",    "L",    "N",    "\321", "N",    "N",
	"\323", "\322", "\324", "\326", "\325", "O",    "O",    "R",    "R",
	"R",    "S",    "S",    "S",    "S",    "T",    "T",    "\332", "\331",
	"\333", "\334", "U",    "U",    "U",    "U",    "U",    "U",    "W",
	"\335", "Y",    "Y",    "Z",    "Z",    "Z",    "\341", "\340", "\342",
	"\344", "\343", "a",    "\345", "a",    "a",    "c",    "c",    "c",
	"c",    "\347", "d",    "\351", "\350", "\352", "\353", "e",    "e",
	"e",    "e",    "g",    "g",    "g",    "g",    "h",    "\355", "\354",
	"\356", "\357", "i",    "i",    "i",    "j",    "k",    "l",    "l",
	"l",    "n",    "\361", "n",    "n",    "\363", "\362",  undef, "\364",
	"\366", "\365", "o",    "o",    "r",    "r",    "r",    "s",    "s",
	"s",    "s",    "t",    "t",    "\372", "\371", "\373", "\374", "u",
	"u",    "u",    "u",    "u",    "u",    "w",    "\375", "y",    "\377",
	"z",    "z",    "z",    "\301", "\242", "\243", undef,  "\245", undef,
	"\247", "\244", "\253", "\260", "\261", "\262", "\263", "\327", "\265",
	"\266", "\267", "\367", "\273", "\274", "\275", "\276", "\277", undef,
	"\306", "\320", "\252", undef,  undef,  undef,  undef,  "\330", undef,
	"\272", "\336", undef,  undef,  undef,  undef,  "\346", "\360", undef,
	undef,  undef,  undef,  undef,  undef,  "\370", undef,  "\337", "\376",
	undef,  undef,  "\241", undef,  undef
	],
    iso8859_9 => [	# ISO8859_9, ISO9241, LATIN5
	undef,  "\300", "\302", "\304", "\303", "A",    "\305", "A",    "A",
	"C",    "C",    "C",    "C",    "\307", "D",    "\311", "\310", "\312",
	"\313", "E",    "E",    "E",    "E",    "G",    "\320", "G",    "G",
	"H",    "\315", "\314", "\316", "\317", "I",    "\335", "I",    "I",
	"J",    "K",    "L",    "L",    "L",    "N",    "\321", "N",    "N",
	"\323", "\322", "\324", "\326", "\325", "O",    "O",    "R",    "R",
	"R",    "S",    "S",    "S",    "\336", "T",    "T",    "\332", "\331",
	"\333", "\334", "U",    "U",    "U",    "U",    "U",    "U",    "W",
	"Y",    "Y",    "Y",    "Z",    "Z",    "Z",    "\341", "\340", "\342",
	"\344", "\343", "a",    "\345", "a",    "a",    "c",    "c",    "c",
	"c",    "\347", "d",    "\351", "\350", "\352", "\353", "e",    "e",
	"e",    "e",    "g",    "g",    "\360", "g",    "h",    "\355", "\354",
	"\356", "\357", "i",    "i",    "i",    "j",    "k",    "l",    "l",
	"l",    "n",    "\361", "n",    "n",    "\363", "\362", undef,  "\364",
	"\366", "\365", "o",    "o",    "r",    "r",    "r",    "s",    "s",
	"s",    "\376", "t",    "t",    "\372", "\371", "\373", "\374", "u",
	"u",    "u",    "u",    "u",    "u",    "w",    "y",    "y",    "\377",
	"z",    "z",    "z",    "\301", "\242", "\243", undef,  "\245", undef,
	"\247", "\244", "\253", "\260", "\261", "\262", "\263", "\327", "\265",
	"\266", "\267", "\367", "\273", "\274", "\275", "\276", "\277", undef,
	"\306", undef,  "\252", undef,  undef,  undef,  undef,  "\330", undef,
	"\272", undef,  undef,  undef,  undef,  undef,  "\346", undef,  undef,
	undef,  "\375", undef,  undef,  undef,  "\370", undef,  "\337", undef,
	undef,  undef,  "\241", undef,  undef
	],
    ocrb      => [	# OCRB
	undef,  "A",    "A",    "A",    "A",    "A",    "A",    "A",    "A",
	"C",    "C",    "C",    "C",    "C",    "D",    "E",    "E",    "E",
	"E",    "E",    "E",    "E",    "E",    "G",    "G",    "G",    "G",
	"H",    "I",    "I",    "I",    "I",    "I",    "I",    "I",    "I",
	"J",    "K",    "L",    "L",    "L",    "N",    "N",    "N",    "N",
	"O",    "O",    "O",    "O",    "O",    "O",    "O",    "R",    "R",
	"R",    "S",    "S",    "S",    "S",    "T",    "T",    "U",    "U",
	"U",    "U",    "U",    "U",    "U",    "U",    "U",    "U",    "W",
	"Y",    "Y",    "Y",    "Z",    "Z",    "Z",    "A",    "A",    "A",
	"A",    "A",    "A",    "A",    "A",    "A",    "C",    "C",    "C",
	"C",    "C",    "D",    "E",    "E",    "E",    "E",    "E",    "E",
	"E",    "E",    "G",    "G",    "G",    "G",    "H",    "I",    "I",
	"I",    "I",    "I",    "I",    "I",    "J",    "K",    "L",    "L",
	"L",    "N",    "N",    "N",    "N",    "O",    "O",    undef,  "O",
	"O",    "O",    "O",    "O",    "R",    "R",    "R",    "S",    "S",
	"S",    "S",    "T",    "T",    "U",    "U",    "U",    "U",    "U",
	"U",    "U",    "U",    "U",    "U",    "W",    "Y",    "Y",    "Y",
	"Z",    "Z",    "Z",    "A",    "",     "",     "",     "",     "",
	"",     "",     "",     "",     "",     "",     "",     "",     "",
	"",     "",     "",     "",     "",     "",     "",     "",     "",
	"AE",   "D",    "IF",   "H",    "",     "L",    "L",    "O",    "OE",
	"IM",   "TI",   "T",    "ENG",  "N",    "KG",   "AE",   "D",    "ETH",
	"H",    "I",    "",     "L",    "L",    "O",    "OE",   "SS",   "TI",
	"T",    "ENG",  "",     "",     ""
	],
    utf8  => [	# Unicode
	undef,
	"\N{LATIN CAPITAL LETTER A WITH GRAVE}",
	"\N{LATIN CAPITAL LETTER A WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER A WITH DIAERESIS}",
	"\N{LATIN CAPITAL LETTER A WITH TILDE}",
	"\N{LATIN CAPITAL LETTER A WITH BREVE}",
	"\N{LATIN CAPITAL LETTER A WITH RING ABOVE}",
	"\N{LATIN CAPITAL LETTER A WITH MACRON}",
	"\N{LATIN CAPITAL LETTER A WITH OGONEK}",
	"\N{LATIN CAPITAL LETTER C WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER C WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER C WITH CARON}",
	"\N{LATIN CAPITAL LETTER C WITH DOT ABOVE}",
	"\N{LATIN CAPITAL LETTER C WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER D WITH CARON}",
	"\N{LATIN CAPITAL LETTER E WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER E WITH GRAVE}",
	"\N{LATIN CAPITAL LETTER E WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER E WITH DIAERESIS}",
	"\N{LATIN CAPITAL LETTER E WITH CARON}",
	"\N{LATIN CAPITAL LETTER E WITH DOT ABOVE}",
	"\N{LATIN CAPITAL LETTER E WITH MACRON}",
	"\N{LATIN CAPITAL LETTER E WITH OGONEK}",
	"\N{LATIN CAPITAL LETTER G WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER G WITH BREVE}",
	"\N{LATIN CAPITAL LETTER G WITH DOT ABOVE}",
	"\N{LATIN CAPITAL LETTER G WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER H WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER I WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER I WITH GRAVE}",
	"\N{LATIN CAPITAL LETTER I WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER I WITH DIAERESIS}",
	"\N{LATIN CAPITAL LETTER I WITH TILDE}",
	"\N{LATIN CAPITAL LETTER I WITH DOT ABOVE}",
	"\N{LATIN CAPITAL LETTER I WITH MACRON}",
	"\N{LATIN CAPITAL LETTER I WITH OGONEK}",
	"\N{LATIN CAPITAL LETTER J WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER K WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER L WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER L WITH CARON}",
	"\N{LATIN CAPITAL LETTER L WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER N WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER N WITH TILDE}",
	"\N{LATIN CAPITAL LETTER N WITH CARON}",
	"\N{LATIN CAPITAL LETTER N WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER O WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER O WITH GRAVE}",
	"\N{LATIN CAPITAL LETTER O WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER O WITH DIAERESIS}",
	"\N{LATIN CAPITAL LETTER O WITH TILDE}",
	"\N{LATIN CAPITAL LETTER O WITH DOUBLE ACUTE}",
	"\N{LATIN CAPITAL LETTER O WITH MACRON}",
	"\N{LATIN CAPITAL LETTER R WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER R WITH CARON}",
	"\N{LATIN CAPITAL LETTER R WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER S WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER S WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER S WITH CARON}",
	"\N{LATIN CAPITAL LETTER S WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER T WITH CARON}",
	"\N{LATIN CAPITAL LETTER T WITH CEDILLA}",
	"\N{LATIN CAPITAL LETTER U WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER U WITH GRAVE}",
	"\N{LATIN CAPITAL LETTER U WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER U WITH DIAERESIS}",
	"\N{LATIN CAPITAL LETTER U WITH TILDE}",
	"\N{LATIN CAPITAL LETTER U WITH BREVE}",
	"\N{LATIN CAPITAL LETTER U WITH DOUBLE ACUTE}",
	"\N{LATIN CAPITAL LETTER U WITH RING ABOVE}",
	"\N{LATIN CAPITAL LETTER U WITH MACRON}",
	"\N{LATIN CAPITAL LETTER U WITH OGONEK}",
	"\N{LATIN CAPITAL LETTER W WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER Y WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER Y WITH CIRCUMFLEX}",
	"\N{LATIN CAPITAL LETTER Y WITH DIAERESIS}",
	"\N{LATIN CAPITAL LETTER Z WITH ACUTE}",
	"\N{LATIN CAPITAL LETTER Z WITH CARON}",
	"\N{LATIN CAPITAL LETTER Z WITH DOT ABOVE}",
	"\N{LATIN SMALL LETTER A WITH ACUTE}",
	"\N{LATIN SMALL LETTER A WITH GRAVE}",
	"\N{LATIN SMALL LETTER A WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER A WITH DIAERESIS}",
	"\N{LATIN SMALL LETTER A WITH TILDE}",
	"\N{LATIN SMALL LETTER A WITH BREVE}",
	"\N{LATIN SMALL LETTER A WITH RING ABOVE}",
	"\N{LATIN SMALL LETTER A WITH MACRON}",
	"\N{LATIN SMALL LETTER A WITH OGONEK}",
	"\N{LATIN SMALL LETTER C WITH ACUTE}",
	"\N{LATIN SMALL LETTER C WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER C WITH CARON}",
	"\N{LATIN SMALL LETTER C WITH DOT ABOVE}",
	"\N{LATIN SMALL LETTER C WITH CEDILLA}",
	"\N{LATIN SMALL LETTER D WITH CARON}",
	"\N{LATIN SMALL LETTER E WITH ACUTE}",
	"\N{LATIN SMALL LETTER E WITH GRAVE}",
	"\N{LATIN SMALL LETTER E WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER E WITH DIAERESIS}",
	"\N{LATIN SMALL LETTER E WITH CARON}",
	"\N{LATIN SMALL LETTER E WITH DOT ABOVE}",
	"\N{LATIN SMALL LETTER E WITH MACRON}",
	"\N{LATIN SMALL LETTER E WITH OGONEK}",
	"\N{LATIN SMALL LETTER G WITH ACUTE}",
	"\N{LATIN SMALL LETTER G WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER G WITH BREVE}",
	"\N{LATIN SMALL LETTER G WITH DOT ABOVE}",
	"\N{LATIN SMALL LETTER H WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER I WITH ACUTE}",
	"\N{LATIN SMALL LETTER I WITH GRAVE}",
	"\N{LATIN SMALL LETTER I WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER I WITH DIAERESIS}",
	"\N{LATIN SMALL LETTER I WITH TILDE}",
	"\N{LATIN SMALL LETTER I WITH MACRON}",
	"\N{LATIN SMALL LETTER I WITH OGONEK}",
	"\N{LATIN SMALL LETTER J WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER K WITH CEDILLA}",
	"\N{LATIN SMALL LETTER L WITH ACUTE}",
	"\N{LATIN SMALL LETTER L WITH CARON}",
	"\N{LATIN SMALL LETTER L WITH CEDILLA}",
	"\N{LATIN SMALL LETTER N WITH ACUTE}",
	"\N{LATIN SMALL LETTER N WITH TILDE}",
	"\N{LATIN SMALL LETTER N WITH CARON}",
	"\N{LATIN SMALL LETTER N WITH CEDILLA}",
	"\N{LATIN SMALL LETTER O WITH ACUTE}",
	"\N{LATIN SMALL LETTER O WITH GRAVE}",
	undef,
	"\N{LATIN SMALL LETTER O WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER O WITH DIAERESIS}",
	"\N{LATIN SMALL LETTER O WITH TILDE}",
	"\N{LATIN SMALL LETTER O WITH DOUBLE ACUTE}",
	"\N{LATIN SMALL LETTER O WITH MACRON}",
	"\N{LATIN SMALL LETTER R WITH ACUTE}",
	"\N{LATIN SMALL LETTER R WITH CARON}",
	"\N{LATIN SMALL LETTER R WITH CEDILLA}",
	"\N{LATIN SMALL LETTER S WITH ACUTE}",
	"\N{LATIN SMALL LETTER S WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER S WITH CARON}",
	"\N{LATIN SMALL LETTER S WITH CEDILLA}",
	"\N{LATIN SMALL LETTER T WITH CARON}",
	"\N{LATIN SMALL LETTER T WITH CEDILLA}",
	"\N{LATIN SMALL LETTER U WITH ACUTE}",
	"\N{LATIN SMALL LETTER U WITH GRAVE}",
	"\N{LATIN SMALL LETTER U WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER U WITH DIAERESIS}",
	"\N{LATIN SMALL LETTER U WITH TILDE}",
	"\N{LATIN SMALL LETTER U WITH BREVE}",
	"\N{LATIN SMALL LETTER U WITH DOUBLE ACUTE}",
	"\N{LATIN SMALL LETTER U WITH RING ABOVE}",
	"\N{LATIN SMALL LETTER U WITH MACRON}",
	"\N{LATIN SMALL LETTER U WITH OGONEK}",
	"\N{LATIN SMALL LETTER W WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER Y WITH ACUTE}",
	"\N{LATIN SMALL LETTER Y WITH CIRCUMFLEX}",
	"\N{LATIN SMALL LETTER Y WITH DIAERESIS}",
	"\N{LATIN SMALL LETTER Z WITH ACUTE}",
	"\N{LATIN SMALL LETTER Z WITH CARON}",
	"\N{LATIN SMALL LETTER Z WITH DOT ABOVE}",
	"\N{LATIN CAPITAL LETTER A WITH ACUTE}",
	"\N{CENT SIGN}",
	"\N{POUND SIGN}",
	"\N{DOLLAR SIGN}",
	"\N{YEN SIGN}",
	"\N{NUMBER SIGN}",
	"\N{SECTION SIGN}",
	"\N{CURRENCY SIGN}",
	"\N{LEFT-POINTING DOUBLE ANGLE QUOTATION MARK}",
	"\N{DEGREE SIGN}",
	"\N{PLUS-MINUS SIGN}",
	"\N{SUPERSCRIPT TWO}",
	"\N{SUPERSCRIPT THREE}",
	"\N{MULTIPLICATION SIGN}",
	"\N{MICRO SIGN}",
	"\N{PILCROW SIGN}",
	"\N{MIDDLE DOT}",
	"\N{DIVISION SIGN}",
	"\N{RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK}",
	"\N{VULGAR FRACTION ONE QUARTER}",
	"\N{VULGAR FRACTION ONE HALF}",
	"\N{VULGAR FRACTION THREE QUARTERS}",
	"\N{INVERTED QUESTION MARK}",
	"\N{OHM SIGN}",
	"\N{LATIN CAPITAL LETTER AE}",
	"\N{LATIN CAPITAL LETTER D WITH STROKE}",
	"\N{FEMININE ORDINAL INDICATOR}",
	"\N{LATIN CAPITAL LETTER H WITH STROKE}",
	"\N{LATIN CAPITAL LIGATURE IJ}",
	"\N{LATIN CAPITAL LETTER L WITH MIDDLE DOT}",
	"\N{LATIN CAPITAL LETTER L WITH STROKE}",
	"\N{LATIN CAPITAL LETTER O WITH STROKE}",
	"\N{LATIN CAPITAL LIGATURE OE}",
	"\N{MASCULINE ORDINAL INDICATOR}",
	"\N{LATIN CAPITAL LETTER THORN}",
	"\N{LATIN CAPITAL LETTER T WITH STROKE}",
	"\N{LATIN CAPITAL LETTER ENG}",
	"\N{LATIN SMALL LETTER N PRECEDED BY APOSTROPHE}",
	"\N{LATIN SMALL LETTER KRA}",
	"\N{LATIN SMALL LETTER AE}",
	"\N{LATIN SMALL LETTER D WITH STROKE}",
	"\N{LATIN SMALL LETTER ETH}",
	"\N{LATIN SMALL LETTER H WITH STROKE}",
	"\N{LATIN SMALL LETTER DOTLESS I}",
	"\N{LATIN SMALL LIGATURE IJ}",
	"\N{LATIN SMALL LETTER L WITH MIDDLE DOT}",
	"\N{LATIN SMALL LETTER L WITH STROKE}",
	"\N{LATIN SMALL LETTER O WITH STROKE}",
	"\N{LATIN SMALL LIGATURE OE}",
	"\N{LATIN SMALL LETTER SHARP S}",
	"\N{LATIN SMALL LETTER THORN}",
	"\N{LATIN SMALL LETTER T WITH STROKE}",
	"\N{LATIN SMALL LETTER ENG}",
	"\N{INVERTED EXCLAMATION MARK}",
	"\N{LATIN SMALL LETTER LAMBDA WITH STROKE}" ] # entry voor DIAC_UNKNOWN
    );

# sub INIT
{
    foreach my $key ("\x80" .. "\xFF") {
	defined $DiacCP500{$key} or ($DiacCP500{$key} = "\xFF");
	}

    foreach my $i (0 .. $#gba_tab) {
	defined $gba_tab[$i][2] or next;
	my $tex = $gba_tab[$i][2];
	$gba_tab[$i][0] == 2 and $tex .= $gba_tab[$i][1];
	$dbd_tab{$tex} = [ $i, $gba_tab[$i][1] ];
	$gba_tab{$gba_tab[$i][3]} = $i;	# Ref by alias
	my $u = $cs{utf8}->[$i] or next;
	$uni_tab{$u} = [ $tex, $i, $gba_tab[$i][3], $u ];
	}
    foreach my $cs (keys %cs) {
	foreach my $i (0 .. $#gba_tab) {
	    defined $cs{$cs}->[$i] or $cs{$cs}->[$i] = $gba_tab[$i][1];
	    }
	}
    } # INIT

sub diac_debug
{
    $debug = $_[0];
    } # diac_debug

# counts extra space used for diac's
sub gbadiaccnt ($)
{
    my $s = shift;
    $s =~ s/[\xC1-\xCF]//g;
    } # gbadiaccnt

# Return length of string after diac's have been resolved
sub diac_length ($)
{
    my $s = shift;
    length ($s) - gbadiaccnt ($s);
    } # diac_length

# diac_substr ("Fran\xCBcois", 1)    => "ran\xCBcois"
# diac_substr ("Fran\xCBcois", 2, 4) => "an\xCBco"
# diac_substr ("Fran\xCBcois", 3, 8) => "n\xCBcois   "
sub diac_substr
{
    my ($s, $f, $l) = @_;
    my @s = ($s =~ m/([\xC1-\xCF]?[^\xC1-\xCF])/g);
    if (defined $l) {
	@s = splice @s, $f, $l;
	$l > @s and push @s, " " x ($l - @s);
	}
    else {
	splice @s, 0, $f;
	}
    join "", @s;
    } # diac_substr

sub diacgba_check ($$)
{
    my ($db, $d) = @_;
    my (@d, %d);

    defined $d  or return @d;

    my $l = length $db;
    $d =~ s/^((..)*)?(  )+$/$1/;	# Force legal pairs
    while ($d =~ s/^(.)(.)//s) {
	my ($idx, $diac) = (ord $1, ord $2);
	my $g = $diac <= $#gba_tab ? $gba_tab[$diac] : [ 0, '?', 0, "?" ];
	if (exists $d{$idx}) {
	    push @d, sprintf "Diac %3d (%s) on pos %2d overloads %3d (%s) on same pos",
		$diac, $g->[3], $idx, $d{$idx}, $gba_tab[$d{$idx}][3];
	    next;
	    }
	if ($diac >= $#gba_tab) {	# includes entry for unsupported
	    push @d, sprintf "Diac %3d on pos %2d unsupported", $diac, $idx;
	    next;
	    }
	if ($idx > $l) {
	    push @d, sprintf "Diac %3d (%s) on pos %2d past stringlength (%3d)",
		$diac, $g->[3], $idx, $l;
	    next;
	    }
	$d{$idx} = $diac;
	}
    foreach my $idx (1 .. $l) {
	my $bc = substr ($db, $idx - 1, 1);
	$dsw[ord $bc] or
	    push @d, sprintf "Illegal base char '%s' (%3d/0x%02X) on pos %2d", $bc, ord $bc, ord $bc, $idx;
    	}
    foreach my $idx (sort { $a <=> $b } keys %d) {
	my $bc = substr ($db, $idx - 1, 1);
	my $gba = $gba_tab[$d{$idx}] or
	    push @d, sprintf "Unknown diac %3d on pos %2d", $d{$idx}, $idx;
	$bc eq $gba->[1] or
	    push @d, sprintf "Base char '%s' for diac %3d (%s) on pos %2d should be '%s'", $bc, $d{$idx}, $gba->[3], $idx, $gba->[1];
	}
    @d;
    } # diacgba_check

# Z     a     i     l     e     n     o     r      +	DB
# 01 4D 02 4E 03 6D 04 C9 05 60 06 77 07 CB 08 82  =	D
# C7 5A C2 61 C8 69 F7 C8 65 C4 6E F9 C2 72		GBA
sub diacgba ($$)
{
    my ($db, $d) = @_;
    my %d;

    defined $db or return "";
    defined $d  or return $db;

    $debug and print STDERR "\N{LEFT-POINTING DOUBLE ANGLE QUOTATION MARK}$db(";
    my $l = length $db;
    $d =~ s/^((..)*)?(  )+$/$1/;	# Force legal pairs
    $d{ord $1} = ord $2 while $d =~ s/^(.)(.)//s;
    foreach my $idx (reverse sort { $a <=> $b } keys %d) {
	$idx > $l and next;	# Safety first
	my $gba = $gba_tab[$d{$idx}];
	$debug and printf STDERR "%d:%3d:%s,", $idx, $d{$idx}, $gba->[3];
	substr ($db, $idx - 1, 1) =
	    $gba->[0] == 1 ? $gba->[2] : "$gba->[2]$gba->[1]";
	}
    $debug and print STDERR "\b)$db\N{RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK}\r\n";
    $db;
    } # diacgba

# C7 5A C2 61 C8 69 F7 C8 65 C4 6E F9 C2 72	   =	GBA
# Z     a     i     l     e     n     o     r      +	DB
# 01 4D 02 4E 03 6D 04 C9 05 60 06 77 07 CB 08 82	D
sub diacdbd ($;$$)
{
    my $gba   = shift;
    my $force = shift || "";
    my $ascii = shift || 0;
    my $warn  = 0;

    my $str   = $gba;
    my $db    = "";
    my $d     = "";

    $gba =~ s/\s+$//;
    $debug and print STDERR "\N{LEFT-POINTING DOUBLE ANGLE QUOTATION MARK}$gba(";
    $force =~ s/(?<=.)!$// and $warn = 1;
    my $rep = $force || "?";
    if ($force ne "1" && $gba =~ s/[\x23\x24\x7B\x7D-\xA0\xA9\xAA\xAC-\xAF\xB9\xBA\xC0\xC9\xD0-\xDF\xE5\xFF]/$rep/g) {
	my $err = "String contains illegal TeleTex character(s)";
	$force  or return (undef, $err);
	$warn  and print STDERR "+++ '$str': $err\n";
	}
    my $re = qr/(?:[] -"%-[_a-z|]+
		  |[\xC1-\xC8\xCA-\xCF][a-zA-Z]
		  |[\xA1-\xA8\xAB\xB0-\xB8\xBB-\xBF\xE0-\xE4\xE6-\xFE]
		  )/x;
    $re = $force eq "1" ? qr/^($re|.)/ : qr/^($re|\Q$rep\E)/;
    while ($gba =~ s/$re//) {
	my $x = $1;
	$debug and print STDERR "$x";
	if (ord ($x) <= 0x7C) {
	    $db .= $x;
	    next;
	    }
	unless (exists $dbd_tab{$x}) {
	    my $err = "Illegal TeleTex sequence '$x'";
	    $force  or return (undef, $err);
	    $warn  and print STDERR "+++ '$str': $err\n";
	    $db .= $x;
	    next;
	    }
	my $xx = $dbd_tab{$x};
	$db .= $xx->[1];
	my $p = length $db;
	$debug and print STDERR "->$p:[$xx->[0],$xx->[1]]";
	$d .= $ascii ? ",$p:$xx->[0]" : pack "CC", $p, $xx->[0];
	$p++;
	} continue {
	    $debug and print STDERR ",";
	    }
    $ascii and $d =~ s/^,//;
    $debug and print STDERR ") =>\n\t('$db' + '$d')\n";
    ($db, $d);
    } # diacdbd

# C7 5A C2 61 C8 69 F7 C8 65 C4 6E F9 C2 72	   =	GBA
# Z     a     i     l     e     n     o     r      +	DB
# 01 4D 02 4E 03 6D 04 C9 05 60 06 77 07 CB 08 82	D
sub diacdbx ($;$)
{
    diacdbd ($_[0], $_[1], 1);
    } # diacdbx

# converts Code Page 500 EBCDIC->ASCII-translation to TELETEX
# returns: 0 if no diacs found
#          1 if diacs found
#          2 if unknown characters and/or diacs found

sub CnvDiacCP500 ($)
{
    my $s = shift;

    ${$s} =~ s/([\x80-\xFF])/$DiacCP500{$1}/g	or return 0;
    ${$s} =~ s/\xFF/#/g				or return 1;
    2;
    } # CnvDiacCP500

# converts iso8859-1 and iso8859-9 to TELETEX
# returns: 0 if no diacs found
#          1 if diacs found
#          2 if unknown characters and/or diacs found

sub CnvDiac8859_1 ($)
{
    my $s = shift;

    ${$s} =~ s/([\x80-\xFF])/$Diac8859_1{$1}/g	or return 0;
    ${$s} =~ s/\xFF/#/g				or return 1;
    2;
    } # CnvDiac8859_1

sub CnvDiac8859_9 ($)
{
    my $s = shift;

    ${$s} =~ s/([\x80-\xFF])/$Diac8859_9{$1}/g	or return 0;
    ${$s} =~ s/\xFF/#/g				or return 1;
    2;
    } # CnvDiac8859_9

sub Tex2cs ($$)
{
    my ($s, $cs) = @_;

    defined $s or return undef;
    $s =~ s/([\xC1-\xC8\xCA\xCb\xCD-\xCF].
	   |[\xA1-\xA8\xAB\xB0-\xB8\xBB-\xBF\xE0-\xE4\xE6-\xFE]
	   )
	  /$cs{$cs}->[$dbd_tab{$1}[0] || 178]	# IQM
	  /gex;
    $s;
    } # Tex2cs

# Strip *ALL* diac's from TELETEX
sub TexNone ($)
{
    Tex2cs ($_[0], "cs7");
    } # TexNone

# converts TELETEX to multinational cs8 (wyse)
sub Texcs8 ($)
{
    Tex2cs ($_[0], "cs8");
    } # Texcs8

# converts TELETEX to OCRB
sub TexOCRB ($)
{
    Tex2cs ($_[0], "ocrb");
    } # TexOCRB

# converts TELETEX to iso8859-1
sub Tex8859_1 ($)
{
    Tex2cs ($_[0], "iso8859_1");
    } # Tex8859_1

# converts TELETEX to iso8859-5
sub Tex8859_9 ($)
{
    Tex2cs ($_[0], "iso8859_9");
    } # Tex8859_9

# converts TELETEX to Unicode
sub TexUnicode ($)
{
    my $uni = Tex2cs ($_[0], "utf8");
    utf8::upgrade ($uni);
    $uni;
    } # TexUnicode

# converts Unicode to TELETEX
sub UnicodeTex ($)
{
    join "", map { exists $uni_tab{$_} ? $uni_tab{$_}[0] : $_ } split //, $_[0];
    } # UnicodeTex

sub DiacLookup ($$)
{
    my ($charset, $char) = @_;
    $charset eq "utf8" && exists $uni_tab{$char} and return $uni_tab{$char};
    foreach my $u ("\n", "_") {			# map l_STROKE => lSTROKE
	substr ($char, 1, 1) =~ s/$u//;
	foreach my $chr ($char, $char."_IDX") {	# map lSTROKE  => lSTROKE_IDX
	    exists $gba_tab{$chr} or next;
	    my $i = $gba_tab{$chr};
	    my $c = $cs{$charset}->[$i];
	    my $t = $gba_tab[$i][2];
	    $gba_tab[$i][0] == 2 and $t .= $gba_tab[$i][1];
	    return [ $t, $i, $chr, defined $c ? $c : "" ];
	    }
	}
    return [ "", 0, undef, "" ];
    } # DiacLookup

1;
