package PROCURA::Format;

$PROCURA::Format::VERSION = 1.21;

use strict;
no strict "refs";
use Data::Dumper;
use Exporter;
use Carp;
use PROCURA::Diac;
use Symbol;
use vars qw(
    @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS
    $VERSION $debug @maand
    $d_sys $D_sys $D_sys_d $D_sys_m $D_sys_j
    $T_sys $C_usr $S_usr
    );

@ISA     = qw(Exporter);

@EXPORT  = qw(
    pr_tijd
    pr_datum
    pr_getal
    pr_snr
    pr_anr
    pr_bank
    pr_pc
    );

@EXPORT_OK = qw(
    $d_sys
    $D_sys_d
    $D_sys_m
    $D_sys_j
    $D_sys
    $T_sys
    $C_usr
    $S_usr
    @maand
    );

%EXPORT_TAGS = (
        subs => [qw(pr_tijd pr_datum pr_getal pr_snr pr_anr pr_bank pr_pc)],
        vars => [qw($d_sys $D_sys_d $D_sys_m $D_sys_j $D_sys $T_sys $C_usr $S_usr @maand)],
        );

@maand = ( "",
    "januari", "februari", "maart", "april", "mei", "juni",
    "juli", "augustus", "september", "oktober", "november", "december"
    );

my @time = localtime (time);
$D_sys_d = $time[3];
$D_sys_m = $time[4] + 1;
$D_sys_j = $time[5] + 1900;
$d_sys   = $D_sys_j * 10000 + $D_sys_m * 100 + $D_sys_d;
$D_sys   = sprintf "%02d-%02d-%04d", $D_sys_d, $D_sys_m, $D_sys_j;
$T_sys   = sprintf "%02d:%02d.%02d", $time[2], $time[1], $time[0];
$C_usr   = $>;
$S_usr   = (getpwuid ($>))[0];

sub debug ($$@)
{
    my ($sub, @vars) = @_;
    my $this = shift @vars;
    my $h    = \%{*$this};

    my $s = $this;
    $s =~ s/^.*=//;
    printf STDERR "%-50.50s (%s)\n",
        $s."  " x $h->{Level}.$sub,
        join (",", @vars);

    $h->{Level}++;
    } # debug

# sub new (CLASS,HANDLE,PAGE_LENGTH,LEFT_MARGIN,TOP_MARGIN,BOTTOM_MARGIN)
sub new ($$$$$)
{
    my $xx = gensym ();
    my $h  = \%{*$xx};

    $debug and $h->{Level} = 0;
    $debug and debug ("new", @_);

    $h->{CLASS}         = shift; # "PROCURA::Format"
    $h->{OPEN_CMD}      = shift;
    $h->{PAGE_LENGTH}   = shift;
    $h->{LEFT_MARGIN}   = shift;
    $h->{TOP_MARGIN}    = shift;
    $h->{BOTTOM_MARGIN} = shift;

    $h->{OPEN_CMD}      =~ m/^\s*[|>]/ or croak "Failed: new('".$h->{OPEN_CMD}."') Command must start with a '|' or '>'";
    $h->{OPEN_CMD}      =~ m/`/       and croak "Failed: new('".$h->{OPEN_CMD}."') No backticks allowed";

    $h->{PAGE_LENGTH}   =~ m/^\d+$/ or croak "PAGE_LENGTH: '".  $h->{PAGE_LENGTH}  ."' is not numeric";
    $h->{LEFT_MARGIN}   =~ m/^\d+$/ or croak "LEFT_MARGIN: '".  $h->{LEFT_MARGIN}  ."' is not numeric";
    $h->{TOP_MARGIN}    =~ m/^\d+$/ or croak "TOP_MARGIN: '".   $h->{TOP_MARGIN}   ."' is not numeric";
    $h->{BOTTOM_MARGIN} =~ m/^\d+$/ or croak "BOTTOM_MARGIN: '".$h->{BOTTOM_MARGIN}."' is not numeric";

    # Default values
    $h->{FORM_FEED}       = "";
    $h->{PAGE_NUMBER}     = 0;
    $h->{LINES_LEFT}      = 0;
    $h->{FOOTER_LENGTH}   = 0;
    $h->{DONT_BREAK}      = 0;
    $h->{HEADER_LENGTH}   = 0;
    $h->{"TOP+HEADER"}    = $h->{TOP_MARGIN} + $h->{HEADER_LENGTH};
    $h->{"FOOTER+BOTTOM"} = $h->{BOTTOM_MARGIN};

    # Vaste variabelen
    $h->{D_sys_d} = $D_sys_d;
    $h->{D_sys_m} = $D_sys_m;
    $h->{D_sys_j} = $D_sys_j;
    $h->{D_sys}   = $D_sys;
    $h->{T_sys}   = $T_sys;
    $h->{C_usr}   = $C_usr;
    $h->{S_usr}   = $S_usr;

    if ($h->{"TOP+HEADER"} + $h->{"FOOTER+BOTTOM"} > $h->{PAGE_LENGTH}) {
        croak ("new: No printable area left");
	}

    open ($xx, $h->{OPEN_CMD}) or croak "Failed: Create filehandle '$h->{OPEN_CMD}'";
    $h->{FD_OPEN} = 1;

    bless $xx, "PROCURA::Format";
    } # new

# sub open (HANDLE)
sub open ($)
{
    my ($this, $value) = @_;
    my $h = \%{*$this};

    $debug and debug ("open", @_);

    $h->{OPEN_CMD} = $value;
    $h->{OPEN_CMD} =~ m/^\s*[|>]/ or croak "Failed: open('$h->{OPEN_CMD}') Command must start with a '|' or '>'";
    $h->{OPEN_CMD} =~ m/`/       and croak "Failed: open('$h->{OPEN_CMD}') No backticks allowed";

    $h->{FD_OPEN} and $this->close ();

    open ($this, $h->{OPEN_CMD}) or croak "Failed: Create filehandle '$h->{OPEN_CMD}'";
    $h->{FD_OPEN} = 1;

    $debug and $h->{Level}--;
    } # open

sub close ()
{
    my ($this) = @_;
    my $h = \%{*$this};

    $debug and debug ("close", @_);

    # Als we niet op het begin van de pagina staan,
    # schrijven we eerst een footer
    $h->{LINES_LEFT} != 0 and $this->write_footer ();

    $h->{FD_OPEN} or croak "File handle already closed!";

    close ($this) or croak "Failed to close file handle $this '$h->{OPEN_CMD}'\n$!";
    $h->{FD_OPEN} = 0;

    $debug and $h->{Level}--;
    } # close

sub DESTROY ()
{
    my ($this) = @_;
    my $h = \%{*$this};

    $debug and debug ("DESTROY", @_);

    $h->{FD_OPEN} and $this->close ($this);

    $debug and $h->{Level}--;
    } # DESTROY

sub bottom_margin
{
    my $this = shift;
    my $h = \%{*$this};

    if (@_) {
        $h->{BOTTOM_MARGIN} = shift;
        $h->{"FOOTER+BOTTOM"} =
            $h->{BOTTOM_MARGIN} + $h->{HEADER_LENGTH};

        if ($h->{"TOP+HEADER"} + $h->{"FOOTER+BOTTOM"} > $h->{PAGE_LENGTH}) {
            croak ("bottom_margin: No printable area left");
	    }
	}
    return ($h->{BOTTOM_MARGIN});
    } # bottom_margin

sub dont_break
{
    my $this = shift;
    my $h = \%{*$this};
    @_ == 0 ? $h->{DONT_BREAK} : ($h->{DONT_BREAK} = shift);
    } # dont_break

sub left_margin
{
    my $this = shift;
    my $h = \%{*$this};
    @_ == 0 ? $h->{LEFT_MARGIN} : ($h->{LEFT_MARGIN} = shift);
    } # left_margin

sub lines_left
{
    my $this = shift;
    my $h = \%{*$this};
    if (@_ > 0) {
        $_[0] > $h->{PAGE_LENGTH} and
            warn "$_[0] is bigger then page length $h->{PAGE_LENGTH}";
	}
    $h->{LINES_LEFT};
    } # lines_left

sub page_length
{
    my $this = shift;
    my $h = \%{*$this};
    if (@_) {
        $h->{PAGE_LENGTH} = shift;

        if ($h->{"TOP+HEADER"} + $h->{"FOOTER+BOTTOM"} > $h->{PAGE_LENGTH}) {
            croak ("page_length: No printable area left");
	    }
	}
    return $h->{PAGE_LENGTH};
    } # page_length

sub page_number
{
    my $this = shift;
    my $h = \%{*$this};
    @_ == 0 ? $h->{PAGE_NUMBER} : ($h->{PAGE_NUMBER} = shift);
    } # page_number

sub top_margin
{
    my $this = shift;
    my $h = \%{*$this};
    if (@_) {
        $h->{TOP_MARGIN} = shift;
        $h->{"TOP+HEADER"} = $h->{TOP_MARGIN} + $h->{HEADER_LENGTH};

        if ($h->{"TOP+HEADER"} + $h->{"FOOTER+BOTTOM"} > $h->{PAGE_LENGTH}) {
            croak ("top_margin: No printable area left");
	    }
	}
    return ($h->{TOP_MARGIN});
    } # top_margin

# sub header (HEADER)
sub header ($)
{
    my ($this, $format) = @_;
    my $h = \%{*$this};

    $debug and debug ("header", @_);

    $h->{FORMAT_HEADER} = \@{"main::".$format};

    $h->{FORMAT_HEADER}[0] =~ m/~/ and croak "No multi-line '~' allowed in header\n";

    # Effe tellen hoe groot de header is
    my @X = split m/\n/ => $h->{FORMAT_HEADER}[0], -1;
    $h->{HEADER_LENGTH} = @X - 1;
    $h->{"TOP+HEADER"}  = $h->{TOP_MARGIN} + $h->{HEADER_LENGTH};

    if ($h->{"TOP+HEADER"} + $h->{"FOOTER+BOTTOM"} > $h->{PAGE_LENGTH}) {
        croak ("header: No printable area left");
	}
    $debug and $h->{Level}--;
    } # header

# sub footer (FOOTER)
sub footer ($)
{
    my ($this, $format) = @_;
    my $h = \%{*$this};

    $debug and debug ("footer", @_);

    $h->{FORMAT_FOOTER} = \@{"main::".$format};

    $h->{FORMAT_FOOTER}[0] =~ m/~/ and croak "No multi-line '~' allowed in footer\n";

    # Effe tellen hoe groot de footer is
    my @X = split m/\n/ => $h->{FORMAT_FOOTER}[0], -1;
    $h->{FOOTER_LENGTH} = @X - 1;
    $h->{"FOOTER+BOTTOM"} = $h->{FOOTER_LENGTH} + $h->{BOTTOM_MARGIN};

    if ($h->{"TOP+HEADER"} + $h->{"FOOTER+BOTTOM"} > $h->{PAGE_LENGTH}) {
        croak ("footer: No printable area left");
	}
    $debug and $h->{Level}--;
    } # footer

sub write_header ()
{
    my ($this) = @_;
    my $h = \%{*$this};

    $debug and debug ("write_header", @_);

    # Als we niet op het begin van de pagina staan,
    # schrijven we eerst een footer, daarna moeten
    # we op een nieuwe pagina staan
    $h->{LINES_LEFT} and $this->write_footer ();
    $h->{LINES_LEFT} = $h->{PAGE_LENGTH};
    $h->{PAGE_NUMBER}++;

    # Schrijf de top-marge
    foreach (1 .. $h->{TOP_MARGIN}) {
        print $this "\n";
        $h->{LINES_LEFT}--;
	}

    # Schrijf de header
    my @fmt = $this->print_format ($h->{FORMAT_HEADER});
    print $this join ("", @fmt);
    $h->{LINES_LEFT} -= @fmt;

    $debug and $h->{Level}--;
    } # write_header

sub form_feed ()
{
    my ($this) = @_;
    my $h = \%{*$this};

    $debug and debug ("write_footer", @_);

    $this->write_footer ();

    $debug and $h->{Level}--;
    } # form_feed

sub write_footer ()
{
    my ($this) = @_;
    my $h = \%{*$this};

    $debug and debug ("write_footer", @_);

    $h->{LINES_LEFT} == 0 and return;

    if ($h->{FOOTER_LENGTH} > 0) {
        # Zorg ervoor dat we op de goede plaats beginnen
        if ($h->{LINES_LEFT} > $h->{'FOOTER+BOTTOM'}) {
            print $this "\n" x ($h->{LINES_LEFT} - $h->{'FOOTER+BOTTOM'});
            $h->{LINES_LEFT} -= $h->{LINES_LEFT} - $h->{'FOOTER+BOTTOM'}
	    }

        # Plaats de footer
        my @fmt = $this->print_format ($h->{FORMAT_FOOTER});
        print $this join ("", @fmt);
        $h->{LINES_LEFT} -= @fmt;
	}

    # Volgende pagina
    if ($h->{FORM_FEED} eq "") {
        # Opvullen met lege regels.
        print $this "\n" x $h->{LINES_LEFT};
        $h->{LINES_LEFT} = 0;
        $debug and $h->{Level}--;
	}
    else {
        # Next form?
        print $this $h->{FORM_FEED};
	}
    } # write_footer

# sub write (FORMAT;NEED)
sub write ($;$)
{
    my ($this, $format, $need) = @_;
    my $h = \%{*$this};

    $debug and debug ("write",@_);

    $h->{FD_OPEN} or croak ("Failed: write('$format') file handle already closed");

    # Prepare format
    my @fmt = $this->print_format (\@{"main::".$format});

    defined $need or $need = 0;

    if ($need < 0) {
        # Verwijder de regels die teveel zijn.
        #splice (@fmt,,,);
	}
    elsif ($h->{DONT_BREAK} && @fmt > $need) {
        # In dit geval is het lengte van het formaat de $need
        $need = @fmt;
	}

    # Past het niet meer, dan schrijven we een footer en header
    if ($h->{LINES_LEFT} < ($h->{"FOOTER+BOTTOM"} + $need)
            && $h->{LINES_LEFT} != 0) {
        $this->write_footer ();
        $this->write_header ();
	}

    foreach my $line (@fmt) {
        $h->{LINES_LEFT} == $h->{"FOOTER+BOTTOM"} and $this->write_footer();
        $h->{LINES_LEFT} == 0                          and $this->write_header();
        print $this $line;
        $h->{LINES_LEFT}--;
	}

    $debug and $h->{Level}--;
    } # write

# sub print_format (FORMAT,CHECK,NEED)
sub print_format ($)
{
    my ($this, $format) = @_;
    my $h = \%{*$this};

    $debug and debug ("print_format", @_);

    # Hier gaan we straks de geformateerde tekst inzetten
    my @formated = ();

    my @format = @$format;
    my $FORMAT = shift @format;
    # sloop de laatste return eraf
    chomp $FORMAT;

    my @fmt = ();

    my (%hash, $idx);
    foreach my $var (@format) {
        $idx++;

        my $ref = ref ($var);

        if (exists $hash{$var}) {
            # We hebben de variabele al een keer opgehaald/berekend
            $fmt[$idx] = \($hash{$var});
	    }
        elsif ($ref eq "REF") {
            my $r = $var;
            if (ref ($$r) eq "CODE") {
                $fmt[$idx] = \($hash{$var} = &{$$r});
		}
            else {
                croak ("Unexpected variabele type " . ref ($$r));
		}
	    }
        elsif ($ref eq "CODE") {
            # Hmm, subroutine. Uitvoeren dan maar. Tevens in cache opslaan en
            # de reference terug geven aan fmt[idx]
            $fmt[$idx] = \($hash{$var} = &{$var});
	    }
        elsif ($ref eq "SCALAR" || $ref eq "") {
            if ($var eq \$%) {
                $fmt[$idx] = \($h->{PAGE_NUMBER});
		}
            elsif (exists $h->{$var}) {
                $fmt[$idx] = \($h->{$var});
		}
            else {
                $fmt[$idx] = \($hash{$var} = ${$var});
		}
	    }
        else {
            die ("Unexpected variabele type ".$ref);
	    }
	}

    $idx = 0;
    FORMAT: foreach my $format_line (split (m/\n/, $FORMAT, -1)) {
        my $FORMAT_LINE;
        my $MULTI_LINE;
        my $EMPTY_LINE;
        my $DOORGAAN;
        my $KEEP_GOING;
        my $IDX;

        if ($format_line =~ s/(^~~|~~$)/  /) {
            $MULTI_LINE = $EMPTY_LINE = 1;
	    }
        elsif ($format_line =~ s/(^~|~$)/ /) {
            $MULTI_LINE = 0;
            $EMPTY_LINE = 1;
	    }
        else {
            $MULTI_LINE = $EMPTY_LINE = 0;
	    }

        $DOORGAAN = 1;
        $IDX      = $idx;

        while ($DOORGAAN) {
            $FORMAT_LINE = $format_line;
            $idx         = $IDX;
            $KEEP_GOING  = 0;

            my ($r, $s, $t, $u, $l1, $l2);

            my $match = '([@^]((<+|>+|[#.]*#|[&.]*&|[|]+))|@)';
            while ($FORMAT_LINE =~ m/$match/o) {
                # Wat gaan we vervangen
                $s = $1;

                # Bepaal de lengte
                $l1 = length $s;

                # Pak de laatste karakter voor het bepalen van de uitlijning
                $t  = substr $s, -1;

                # Pak de eerste karakter voor het bepalen van de uitlijning
                $u  = substr $s, 0, 1;

                # Zoek de variabele op
                $idx++;
                $r = ${$fmt[$idx]};

                if ($r eq "") {
                    # Hmm, bestaat niet, dan opvullen met spaties
                    $r = " " x $l1;
		    }
                elsif ($t =~ m/^[@<>|]/) {
                    # Bepaal de lengte van de variabele
                    $l2 = length ($r) - gbadiaccnt ($r);

                    if ($l1 < $l2) {
                        # Te lang, effe bijknippen

                        if ($u eq '@') {
                            # Probeer zoveel mogelijk erin te proppen
                            $r  = diac_substr ($r, 0, $l1);
			    }
                        elsif ($u eq "^") {
                            # Knip er 1 regel af, en bewaar de rest.
                            my @f = free_fold ($r, $l1);

                            # Regel 1
                            $r = shift @f;

                            # De rest weer opslaan
                            ${$fmt[$idx]} = shift @f;

                            # Ze vlaggetje dat we moeten doorgaan in het geval
                            # van een multi-line
                            $KEEP_GOING = 1;

                            $l2 = length ($r) - gbadiaccnt ($r);
                            my $l = $l1 - $l2;
                            if ($l>0) {
                                # Hmm, regel niet geheel gevuld, ff bijvullen
                                # met spaties.
                                if ($t eq ">") {
                                    # Rechts, dus links aanvullen
                                    $r = (" " x $l) . $r;
				    }
                                elsif ($t eq "|") {
                                    # Center, dus links & rechts aanvullen
                                    $r = (" " x int($l / 2)) . $r;
                                    $r .= " " x ($l - int($l / 2));
				    }
                                else {
                                    # Links, dus rechts aanvullen
                                    $r .= " " x $l;
				    }
				}
			    }
                        else {
                            croak ("Something wrong with format '$format_line'");
			    }
			}
                    elsif ($l1 > $l2) {
                        # Hmm, we hebben nog wat ruimte over. Ff kijken aan
                        # welke kan de spaties moeten. En ze er aanplakken.
                        my $l = $l1 - $l2;
                        if ($t eq ">") {
                            # Rechts, dus links aanvullen
                            $r = (" " x $l) . $r;
			    }
                        elsif ($t eq "|") {
                            # Center, dus links & rechts aanvullen
                            $r = (" " x int($l / 2)) . $r;
                            $r .= " " x ($l - int($l / 2));
			    }
                        else {
                            # Links, dus rechts aanvullen
                            $r .= " " x $l;
			    }

                        if ($u eq "^") {
                            # Variabele poetsen.
                            ${$fmt[$idx]}="";
			    }
			}
                    else {
                        if ($u eq "^") {
                            # Variabele poetsen.
                            ${$fmt[$idx]}="";
			    }
			}
		    }
                elsif ($t eq "#") {
                    # Hmm, verwachten een getalletje
                    # Met of zonder punt?
                    if ($s =~ m/^(@#*)\.(#+)$/) {
                        $r = sprintf ("%".$l1.".".length ($2)."f", $r);
                        $r =~ s/\./,/;
			}
                    else {
                        $r = sprintf ("%".$l1.".0f", $r);
			}

                    # Helaas nodig, omdat bijvoorbeeld "%5.0f" de string niet
                    # limiteer tot vijf karakters.
                    $r = substr $r, 0, $l1;
		    }
                elsif ($t eq '&') {
                    # Hmm, verwachten een getalletje
                    # Met of zonder punt?
                    if ($s =~ m/^(@&*)\.(&+)$/) {
                        $r = sprintf ("%0".$l1.".".length($2)."f", $r);
                        $r =~ s/\./,/;
			}
                    else {
                        $r = sprintf ("%0".$l1.".0f", $r);
			}

                    # Helaas nodig, omdat bijvoorbeeld "%5.0f" de string niet
                    # limiteer tot vijf karakters.
                    $r = substr $r, 0, $l1;
		    }

                # Vervang het veld door onze tekst.
                $FORMAT_LINE =~ s/$match/$r/;
		}

            # Zo alle velden zijn vervangen.

            if ($h->{LEFT_MARGIN}>0) {
                # Linker kantlijn? Dan deze ff toevoegen aan de string.
                $FORMAT_LINE = " " x $h->{LEFT_MARGIN} . $FORMAT_LINE
		}

            # Multi-line en niet niet alle variabelen leeg, dan doorgaan.
            $DOORGAAN = $MULTI_LINE && $KEEP_GOING;

            # Geen lege regel en alleen maar spaties, dan volgende regel
            $EMPTY_LINE && $FORMAT_LINE =~ m/^\s*$/ && next;

            # Verwijder spaties aan einde van regel
            $FORMAT_LINE =~ s/\s+$//;

            # Plak de regel aan het totaal.
            push (@formated, $FORMAT_LINE."\n");
	    }
	}
    $debug and $h->{Level}--;
    return (@formated);
    } # print_format

################ Algemene functies ############

# sub pr_tijd ($)
sub pr_tijd ($)
{
    my ($tijd) = @_;

    $tijd > 0 or return "        ";

    $tijd = "0" x (6 - length ($tijd)) . $tijd;
    $tijd =~ s/(.{2})(.{2})(.{2})$/$1:$2.$3/;
    return $tijd;
    } # pr_tijd

# sub pr_datum ($;$)
sub pr_datum ($;$)
{
    my ($datum, $voluit) = @_;

    $datum > 0 or return "          ";

    $datum .= "0" x (8 - length $datum);
    my ($jaar, $maand, $dag) = unpack "A4A2A2", $datum;
    if ($voluit) {
	$datum = ($dag + 0)." ".$maand[$maand]." ".($jaar + 0);
	}
    else {
	$datum = $dag."-".$maand."-".$jaar;
	}
    return $datum;
    } # pr_datum

# sub pr_getal ($;$)
sub pr_getal ($;$)
{
    my ($getal, $precision) = @_;

    $getal =~ s/,/./;
    if ($precision > 0) {
        $getal = sprintf "%.".$precision."f", $getal;
        $getal =~ s/\./,/;
	}
    else {
        $getal = sprintf "%.0f", $getal;
	}

    1 while ( $getal =~ s/^(.*\d)(\d{3})/$1.$2/ );

    return $getal;
    } # pr_getal

# sub pr_snr ($)
sub pr_snr ($)
{
    my ($snr) = @_;

    $snr > 0 or return "";

    $snr = " " x (9 - length $snr) . $snr;
    $snr =~ s/^(.{4})(.{2})(.{3})$/$1.$2.$3/;
    return $snr;
    } # pr_snr

# sub pr_anr ($)
sub pr_anr ($)
{
    my ($anr) = @_;

    $anr > 0 or return "";

    $anr = " " x (10 - length $anr) . $anr;
    $anr =~ s/^(.{3})(.{4})(.{3})$/$1.$2.$3/;
    return $anr;
    } # pr_anr

# sub pr_bank ($)
sub pr_bank ($)
{
    my ($bank) = @_;

    $bank > 0 or return "";

    $bank = "0" x (10 - length $bank) . $bank;
    $bank =~ s/^(.*)(.{2})(.{3})$/$1.$2.$3/;
    return $bank;
    } # pr_bank

# sub pr_pc ($)
sub pr_pc ($)
{
    my ($pc) = @_;

    $pc = " " x (6 - length $pc) . $pc;
    $pc =~ s/^(.{4})(.{2})$/$1 $2/;
    return $pc;
    } # pr_pc


# -----------------
# Internal use only
sub free_fold ($$)
{
    my ($s, $l) = @_;

    if (length ($s) - gbadiaccnt ($s) < $l) {
        return $s;
	}

    my @z = ();

    my $z = diac_substr ($s, 0, $l + 1);
    $s = diac_substr ($s, $l + 1);
    if ($z =~ s/^(.*)\n(.*?)$//) {
        # Breek af op de new-line
        my ($a,$b) = ($1,$2.$s);
        push @z, $a;
        push @z, $b;
	}
    elsif ($z =~ s/^(.*)\s(.*?)$//) {
        # Breek af op de laatste whitespace.
        my ($a,$b) = ($1,$2.$s);
        push @z, $a;
        push @z, $b;
	}
    else {
        # No white-space? Dan nemen we alles mee
        $s = diac_substr ($z, $l, 1) . $s;
        $z = diac_substr ($z, 0, $l);
        push @z, $z;
        push @z, $s;
	}

    # Verwijder eventueele achtergebleven white-space
    $z[0] =~ s/\s+$//;
    return (@z);
    } # free_fold

1;
__END__

################ Documentation ################

=head1 NAME

PROCURA::Format - A la perl-format maar dan met diacrieten

=head1 SYNOPSIS

METHODES
    close ($)
    footer ($)
    header ($)
    new ($$$$$)
    open ($)
    write ($;$)
    write_footer ()
    form_feed ()
    write_header ($)
    dont_break ($)
    form_feed ($)
    lines_left ($)
    page_number ($)

SUBROUTINES
    pr_tijd ($)        153137      -> 15:31.37
    pr_datum ($;$)     19990628    -> 28-06-1999 of "28 juni 1999"
    pr_getal ($;$)     1083,283    -> 1.083,28
    pr_snr ($)         187272414   -> 1872.72.414
    pr_anr ($)         4747474747  -> 474.7474.474
    pr_bank ($)        362592187   -> 3625.92.187
    pr_pc ($)          1824KR      -> 1824 KR

=head1 DESCRIPTION

De PROCURA::Format module implementeerd een format ( lijkt op die van perl )
die ook diacrieten ondersteund. Om dit voor elkaar te krijgen, is er een
speciale interface gemaakt.

=over 2

=item B<new> (lp_cmd, page length, left margin, top margin, bottom margin)

Open het uitvoer-kanaal en stel de paginagrootte in.

=item B<open> (commando)

Sluit eventuele huidig uitvoer-kanaal en open een nieuwe.

=item B<header> (FORMAT_NAME)

Stel het formaat van de header.

=item B<footer> (FORMAT_NAME)

Stel het formaat van de footer.

=item B<write_header> ()

Schrijf een header-format naar de opgegeven uitvoer. De huidige pagina
wordt eerst beeindigd met een footer als deze er nog niet staat.

=item B<write_footer> ()

Schrijf een footer-format naar de opgegeven uitvoer, wanneer deze zich
nog niet op de einde van de pagina bevindt.

=item B<form_feed> ()

Eigenlijk hetzelfde als write_footer ().

=item B<write> (FORMAT_NAME, $need)

Schrijf de opgegeven format naar de opgegeven uitvoer. Het tweede argument is
optioneel. Het geeft aan hoeveel regels er minimaal vrij moeten zijn op de
pagina. Wanneer dit niet het geval is wordt er doorgesprongen naar de volgende
pagina. Wanneer je een negatief getal opgeeft, zal het opgegeven aantal regel
gebruikt worden. Niet meer en niet minder. Dit is vooral makkelijk wanneer wil
voorkomen dat een variabel format uit de hand loopt.

=item B<dont_break> (0/1)

Wanneer gezet, wordt het formaat doorgeschoven naar de volgende
pagina, wanneer het niet op de huidige pagina past.

=item B<form_feed> (PRINTER-CODES)

Printer-code voor volgende pagina

=item B<lines_left> (REGELS)

Regels over op huidige pagina

=item B<page_number> (PAGINANUMMER)

Huidige paginanummer

=item B<@FORMAT> ___

Het format is een list(array) waarvan de eerste entry het formaat bevat. Deze is
hetzelfde als in perl-format, maar met de uitzondering dat de variabelen er
niet tussen staat. Na de eerste entry volgen de references naar de variabelen
of een subroutine die een string terug geeft die in het format gebruikt worden.
Ze moeten in de zelfde volgorde worden opgegeven als in het format. Dit houdt
in als je een variabele twee maal gebruikt in een format je deze ook twee maal
moet worden opgegeven. Verder kunnen er nog een aantal vaste variabelen opgegeven
worden.

 @<<<<<<<<<<<        Links uitgelijnd
 @>>>>>>>>>>>        Rechts uitgelijnd
 @|||||||||||        Gecentreerd
 @########.##        Numeriek eventueel met punt/comma
 @&&&&&&&&.&&        Numeriek met voorloopnullen en
                     eventueel met punt/comma
 ^<<<<<<<<<<<        Links uitgelijnd, de verbruikte tekst
                     wordt van de variabele afgehaald
 ^>>>>>>>>>>>        Rechts uitgelijnd, de verbruikte tekst
                     wordt van de variabele afgehaald


=item B<VARIABELEN> ___

 "D_sys"            Huidige datum "DD-MM-JJJJ"
 "D_sys_d"          Huidig jaar   "DD"
 "D_sys_m"          Huidig maand  "MM"
 "D_sys_j"          Huidig jaar   "JJJJ"
 "T_sys"            Huidige tijd  "UU:MM.SS"

 "C_usr"            De effectieve uid ($>)
 "S_usr"            De naam uit 'B</etc/passwd>'

=back

=head1 EXAMPLE

 #!/usr/bin/perl
 use PROCURA::Format;

 $TOP[0] = <<EOF;
 @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 --------------------------------------------
 EOF

 $TOP[1] = sub { "Current time: " . scalar localtime time };

 $BOT[0] = <<EOF;
 --------------------------------------------
 Voetje                                   @>>
 EOF
 push (@BOT,\$%);

 $FMT[0] = <<EOF;
 Links/Rechts: @<<<<<<<<<<<< @>>>>>>>>>>>>>>>
 Center      : @|||||||||||||||||||||||||||||
 Getal       : @##########################.##
 Voorloopnul : @&&&&&&&&&&&&&&&&&&&&&&&&&&.&&
 EOF

 $FMT[1] = \$links;
 $FMT[2] = \$rechts;
 $FMT[3] = \$center;
 $FMT[4] = \$getal;
 $FMT[5] = \$voorloopnul;

 $prn = new PROCURA::Format (">x", 10, 5, 0, 0, 0);
 $prn->header ("TOP");
 $prn->footer ("BOT");

 # Wanneer het formaat niet op de huidige pagina past, wordt er naar de
 # volgende pagina gesprongen.
 $prn->{DONT_BREAK} = 1;

 $links  = "links";
 $rechts = "rechts";
 $center = "center";
 $getal  = 123,25;
 $voorloopnul = 123,25;

 $prn->write ("FMT");
 $prn->close ();

 __END__

 Dit geeft de volgende uitvoer:

 +-------------------------------------------------+
 |     Current time: Fri Jul  2 12:02:45 1999      |
 |     --------------------------------------------|
 |     Links/Rechts: links                   rechts|
 |     Center      :             center            |
 |     Getal       :                         123,00|
 |     Voorloopnul : 000000000000000000000000123,00|
 |                                                 |
 |                                                 |
 |     --------------------------------------------|
 |     Voetje                                     1|
 +-------------------------------------------------+

=cut
