package PROCURA::RPT;

$PROCURA::RPT::VERSION = 4.10;

=head1 NAME

RPT - supply perl-RPT interface

=head1 SYNOPSIS

 use PROCURA::RPT;

 # need ()
 need 5;

 # initials (), initials0 (), initials2 ()
 $e010210 = "Gert Erik Henry Merijn Dawn-Merel IJsbrand";
 $e010211 = initials  ($e010210);	# "G.E.H.M.D.IJ."
 $e010212 = initials0 ($e010210);	# "GEHMDIJ"
 $e010213 = initials2 ($e010210);	# "Gert E.H.M.D.IJ."
 $e010214 = initials3 ($e010210, 25);	# "Gert Erik Henry M.D.IJ."

 # ChkAnum ()
 $e010110 = "1010101025";
 if (ChkAnum ("1010101025") || ChkAnum ("101.0101.025") ||
     ChkAnum (1010101025.)  || ChkAnum ("101| 101|025") ||
     ChkAnum (101, 101, 25)) die;
 ($anr, $a1, $a2, $a3) = ChkAnum ($e010110);
 $a1 < 100 and die "$e010110: bad Anum ($a1)";	# $a1 = failure code

 @anrs = parse_anr (@ARGV);             # Allow several formats,
                                        # including files and -
 # ChkBank ()
 $giro = "2323712";
 if (ChkBank ("2323712") || ChkBank ("23.23.712")  ||
     ChkBank (2323712.)  || ChkBank ("23| 23|712") ||
     ChkBank (23, 23, 712)) die;
 ($bank, $g1, $g2, $g3) = ChkBank ($giro);
 $bank eq "" and die "$giro: bad Bank ($g1)";	# $g1 = failure code

 # free_fold ()
 @ff = free_fold ("Gert Han Henry Merijn Francois", 20);
 @ff = free_fold (\@ff, 29);			# was free_fold2 ()

=head1 DESCRIPTION

=over 4

=item B<need (>S<lines>B<);>

Check if there are enough lines left on the current page, if not,
force a page break.

=item S<string = >B<initials (>S<string>B<);>

Return the initials of a string, separated by dots.

=item S<string = >B<initials0 (>S<string>B<);>

Return the initials of a string, without separation.

=item S<string = >B<initials2 (>S<string>B<);>

Return the first full word of the string followed by the initials
of the rest of the words in the string, separated by dots.

=item S<string = >B<initials3 (>S<string, max_length>B<);>

Convert as many words as necassary from the tail of the string
to initials to fit the resulting string in max_length characters.

=item S<bool = >B<ChkAnum (>S<string>B<)>;

=item S<bool = >B<ChkAnum (>S<number [, number, number]>B<)>;

=item S<list = >B<ChkAnum (>S<string>B<)>;

=item S<list = >B<ChkAnum (>S<number [, number, number]>B<)>;

Check if the A-number, passed as argument(s), is valid. If the
argument is a string, it may optionally be separated with non-
numerics into a group of three numbers.

If the return value is evaluated as a scalar, B<ChkAnum ()> returns 0
if the A-number is valid, 1 if it is not. If the return values
is evaluated in an array context: [0]: the full A-number as a
string, [1..3]: the A-number broken into a1, a2 and a3. If a1
is less than 101, the A-number was not valid.

=item S<list = >B<parse_anum (>S<anum [, ...]>B<)>;

=item S<anum = >S<string> | S<file> | S<'-'>

Parse the arguments as A-numbers. Each argument can be either a
string formatted in any way B<ChkAnum ()> understands, or a file
that contains those strings, or a single minus, indicating that
these strings are read from STDIN (or a pipe).

B<parse_anum ()> returns nothing in case any of the arguments
was invalid, or yielded an invalid A-number.

If all strings parsed resulted in a valid A-number, B<parse_anum ()>
returns a list of anonymous lists, each containing the list
context return of B<ChkAnum ()>.

=back

=head1 BUGS

Far from complete

=cut

### ###########################################################################

use Exporter;
use vars qw(@ISA @EXPORT);

@ISA	 = qw(Exporter);
@EXPORT	 = qw(rpt_debug
	      need
	      initials initials0 initials2 initials3
	      free_fold
	      ChkAnum parse_anum
	      ChkBank
	      dbg);

use PROCURA::Diac;
use strict;

use vars qw($debug);
$debug   = 0;

sub rpt_debug
{
    $debug = $_[0];
    } # rpt_debug

# dbg (3, "Whatever: %s (%d\n", $msg, $status);
sub dbg ($$;@)
{
    my ($lvl, $fmt, @args) = @_;
    $lvl <= $main::opt_v and printf STDERR $fmt, @args;
    } # dbg

# need 3
sub need ($)
{
    $- < shift and ($- = 0);
    } # need

# do_initials ("Gert Han Henry Merijn Dawn-Merel IJsbrand", "")
#	=> GHHMDIJ
sub do_initials ($$)
{
    my ($s, $sep) = @_;

    my %di = (		# Keep these Double initials together
	IJ	=> "0",
	);

    foreach my $di (keys %di) {
	${$s} =~ s/\b$di/$di{$di}/g;
	}
    my $x = ${$s} =~ s/\b(\S)\S*\s*/$1$sep/g;
    foreach my $di (keys %di) {
	${$s} =~ s/$di{$di}/$di/g;
	}
    $x;
    } # do_initials

# initials ("Gert Han Henry Merijn") => G.H.H.M.
sub initials
{
    my $s = shift;
    # $s =~ s/\b(\S)\S*\s*/$1./g;
    do_initials (\$s, ".");
    "$s";
    } # initials

# initials0 ("Gert Han Henry Merijn") => GHHM
sub initials0
{
    my $s = shift;
    #$s =~ s/\b(\S)\S*\s*/$1/g;
    do_initials (\$s, "");
    "$s";
    } # initials0

# initials2 ("Gert Han Henry Merijn") => Gert H.H.M.
sub initials2
{
    my $s = shift;
    my $r;

    ($r, $s) = ($s =~ m/^(\S+)\s*(.*)/);
    do_initials (\$s, ".") ? "$r $s" : "$r";
    } # initials2

# Last names to initials until fit in length
# ("Gert Han Henry Merijn", 15) => "Gert Han H.M."
sub initials3 ($$)
{
    my ($s, $l) = @_;
    while (diac_length ($s) > $l) {
	$s =~ s/\s+([-\w]+)(?:\s+(\S+\.))?$// or last;
	$s .= " " . initials ($1) . ($2 || "");
	}
    "$s";
    } # initials3

# free_fold ("Gert Han Henry Merijn Francois", 20) =>
#	("Gert Han Henry", "Francois Merijn")
# free_fold (["Gert", "Han", "Henry", "Merijn", "Francois"], 20) =>
#	("Gert Han Henry", "Francois Merijn")
sub free_fold
{
    my ($s, $l) = @_;
    my (@s, @l);

    ref $s and ($s = join " ", @{$s});
    length ($s) < $l and return $s;

    foreach my $z (split m/\s+/, $s) {
	push @s, $z;
	push @l, length ($z) - gbadiaccnt ($z);
	}
    my @z = ();
    while (@s) {
	# A single word is too large to fit, break
	if ($l[0] > $l) {
	    push @z, diac_substr ($s[0], 0, $l);
	    $s[0] = diac_substr ($s[0], $l);
	    $l[0] -= $l;
	    next;
	    }
	push @z, shift @s;
	my $z = shift @l;
	while (@s && $z + 1 + $l[0] <= $l) {
	    $z[-1] .= " " . shift @s;
	    $z += 1 + shift @l;
	    }
	}
    @z;
    } # free_fold

my %anr_err = (
    0	=> "OK",
    1	=> "Ongeldig formaat",	# non-digits or not three groups of digits
    2	=> "Ongeldige reeks",	# outside 101..989, 101..9898, 101..989
    3	=> "Ongeldige cijfercombinatie", # adjacent digits should differ
    4	=> "Ongeldige sum-11",	# Sum of digits must be multiple of 11
    5	=> "Ongeldige prod-11",	# Prod of digits must be multiple of 11
    );
# 1010101025, 101.0101.025, 101| 101| 25, (101, 101, 25), 1010101025.
sub ChkAnum
{
    my @a = @_;
    my $a;

    ($a = join ".", @_) =~ s/\s+//g;		# All spacing presumed legal
    @a = split m/(\D+)/, $a;			# Non-digits are separators

    if (scalar @a == 1) {			# No separators
	$a[0] =~ m/^(\d{3})(\d{4})(\d{3})$/ or	#  ... thus ten digits
	    return wantarray ? ("", 1, -1, -1) : 1;
	@a = ($1, ".", $2, ".", $3);
	}
    elsif (scalar @a == 5) {			# Two separators
	$a[1] =~ m/^[\.\|]$/ or			#  ... which should be . or |
	    return wantarray ? ("", 1, -1, -1) : 1;
	$a[1] eq $a[3]       or			#  ... and should be equal
	    return wantarray ? ("", 1, -1, -1) : 1;
	}
    else {
	return wantarray ? ("", 1, -1, -1) : 1;	# bad arguments
	}

    splice @a, 3, 1;
    splice @a, 1, 1;
    $_ += 0 for @a;	# Numify
    ($a[0] <= 100 or $a[0] >=  990 or
     $a[1] <= 100 or $a[1] >= 9900 or
     $a[2] <   10 or $a[2] >=  990) and
	return wantarray ? ("", 2, -1, -1) : 2;

    $a = sprintf "%3d%04d%03d", @a;

    my $t = 0;
    my $p = 0;
    my @d = split m//, $a;
    foreach my $i (0 .. 9) {
	$i and $d[$i] == $d[$i - 1] and
	    return wantarray ? ("", 3, -1, -1) : 3;
	$t += $d[$i];
	$p += $d[$i] * (1 << $i);
	}
    $t % 11 and
	return wantarray ? ("", 4, -1, -1) : 4;
    $p % 11 and
	return wantarray ? ("", 5, -1, -1) : 5;

    return wantarray ? ($a, @a) : 0;		# Passed
    } # ChkAnum

# @anrs = parse_anum (@ARGV);
# args are <file>, '-', or command line arguments
sub parse_anum
{
    my @anr;
    my $chk = sub {
	chomp (my $a = $_[0]);
	my @a = ChkAnum ($a);
	if ($a[1] <= 100) {
	    warn "Anr '$_[0]' uit $_[1]: $anr_err{$a[1]}\n";
	    return 0;
	    }
	push @anr, [ @a ];
	}; # chk

    for (@_) {
	if ($_ eq "-") {
	    while (<STDIN>) {
		$chk->($_, "invoer") or return;
		}
	    next;
	    }
	if (-f $_) {
	    open my $f, "< $_" or die "$_: $!\n";
	    my $F = $_;
	    while (<$f>) {
		$chk->($_, "bestand '$F'") or return;
		}
	    close $f;
	    next;
	    }
	$chk->($_, "argument") or return;
	}
    @anr;
    } # parse_anum

# 2323712, 23.23.712, 23| 23| 712, (23, 23, 712), 2323712.
sub ChkBank
{
    my @a = @_;
    my $a;

    ($a = join ".", @_) =~ s/\s+//g;		# All spacing presumed legal
    @a = split m/(\D+)/, $a;			# Non-digits are separators

    if (scalar @a == 1) {			# No separators
	$a[0] =~ m/^(\d{4,5})(\d{2})(\d{3})$/ or#  ... thus nine or ten digits
	    return wantarray ? ("", 1, -1, -1) : 1;
	@a = ($1, ".", $2, ".", $3);
	}
    elsif (@a == 5 || @a == 7) {		# Two or three separators
	$a[1] =~ m/^[\.\|]$/ or			#  ... which should be . or |
	    return wantarray ? ("", 1, -1, -1) : 1;
	$a[1] eq $a[3]       or			#  ... and should be equal
	    return wantarray ? ("", 1, -1, -1) : 1;
	if (@a == 7) {
	    $a[1] eq $a[5] or			#  ... and should be equal
		return wantarray ? ("", 1, -1, -1) : 1;
	    $a[2] > 99 and
		return wantarray ? ("", 2, -1, -1) : 2;
	    splice @a, 0, 3, 100 * $a[0] + $a[2];
	    }
	}
    else {
	return wantarray ? ("", 1, -1, -1) : 1;	# bad arguments
	}

    splice @a, 3, 1;
    splice @a, 1, 1;
    $a[0] > 99999 || $a[0] <= 999 ||
    $a[1] >    99 ||
    $a[2] >   999 and
	return wantarray ? ("", 2, -1, -1) : 2;

    $a = sprintf "%05d%02d%03d", @a;

    my $t = 0;
    my $p = 0;
    my @d = split m//, $a;

    (10 * $d[9] + 9 * $d[8] + 8 * $d[7] + 7 * $d[6] + 6 * $d[5] +
      5 * $d[4] + 4 * $d[3] + 3 * $d[2] + 2 * $d[1] + 1 * $d[0]) % 11 == 0 and
	return wantarray ? ($a, @a) : 0;		# Passed

    return wantarray ? ("", 3, -1, -1) : 3;
    } # ChkBank

1;
