#!/pro/bin/perl

package PROCURA::XML::Entities;

=head1 NAME

PROCURA::XML::Entities - Encode or decode strings with XML 1.0 entities

=head1 SYNOPSIS

 use PROCURA::XML::Entities;

 my $a = "Read &quot;perlre&quot; for explanation of &apos;&amp;&apos;";
 my $b = decode_entities ($a);

 # $b will now be q{Read "perlre" for explanation of '&'}

 $c = encode_entities ($b);

 # $c should be the same as $a

=head1 DESCRIPTION

This module deals with encoding and decoding of strings with XML 1.0
character entities.  The module provides the following functions:

=over 4

=item decode_entities ($string, ...)

This routine replaces XML entities found in the $string with the
corresponding ASCII character. Unrecognized entities are left alone.

If called in void context the arguments are decoded in-place.

This routine is exported by default.

=item encode_entities ($string)

This routine replaces unsafe characters in $string with their entity
representation. Only the characters that need encoding in the XML 1.0
specs are encoded: C<< ' >>, C<< " >>, C<< & >>, C<< < >>, and C<< > >>.

This routine is exported by default.

=back

All these routines modify the string passed as the first argument, if
called in a void context.  In scalar and array contexts, the encoded or
decoded string is returned (without changing the input string).

If you prefer not to import these routines into your namespace, you can
call them as:

  use PROCURA::XML::Entities ();
  $decoded = PROCURA::XML::Entities::decode ($a);
  $encoded = PROCURA::XML::Entities::encode ($a);

The module can also export the %char2entity and the %entity2char
hashes, which contain the mapping from all characters to the
corresponding entities (and vice versa, respectively).

=head1 AUTHOR

H.Merijn Brand for PROCURA B.V. <h.m.brand@procura.nl>

=head1 COPYRIGHT

Copyright 2005-2005 H.Merijn Brand. All rights reserved.

Structure and main setup copied from Gisle Aas' HTML::Entities

This library is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut

use strict;
use warnings;

use vars qw( @EXPORT @EXPORT_OK $VERSION );
use vars qw( %entity2char %char2entity );

use base "Exporter";

@EXPORT    = qw( encode_entities decode_entities );
@EXPORT_OK = qw( %entity2char %char2entity );

$VERSION = "0.02";
sub Version { $VERSION; }

%entity2char = (
    "amp"	=> "&",  # ampersand 
    "gt"	=> ">",  # greater than
    "lt"	=> "<",  # less than
    "quot"	=> '"',  # double quote
    "apos"	=> "'",  # single quote
    );

# Make the opposite mapping
while (my ($entity, $char) = each %entity2char) {
    $char2entity{$char} = "&$entity;";
    }
#delete $char2entity{"'"};  # only one-way decoding
@entity2char{qw( 38 60 62 34 39 )} = qw( & < > " ' ); # Numeric equivs of &<>"'

sub encode_entities (@)
{
    my $ref;
    if (defined wantarray) {
	my $x = $_[0];
	$ref = \$x;     # copy
	}
    else {
	$ref = \$_[0];  # modify in-place
	}

    # Encode '<', '&', '>', '"', and "'"
    $$ref =~ s/([&<>'"])/$char2entity{$1}/ge;
    $$ref;
    } # encode_entities

sub decode_entities (@)
{
    my $array;
    if (defined wantarray) {
	$array = [@_]; # copy
	}
    else {
	$array = \@_;  # modify in-place
	}
    my $c;
    for (@$array) {
	s/(&(\w+);?)/$entity2char{$2} || $1/eg;
	s/(&\#(\d+);?)/$2 < 256 ? chr($2) : $1/eg;
	s/(&\#[xX]([0-9a-fA-F]+);?)/$c = hex($2); $c < 256 ? chr($c) : $1/eg;
	}
    wantarray ? @$array : $array->[0];
    } # decode_entities 

# Set up aliases
*encode = \&encode_entities;
*decode = \&decode_entities;

1;
