#package PROCURA::getopt;

$PROCURA::getopt::VERSION = 4.05;

=head1 NAME

getopt - interpret optional parameters

=head1 SYNOPSIS

    use PROCURA::getopt;

    $opt_a = 0;
    $opt_b = "";
    $opt_c = undef;
    @opt_d = ();
    $parsed = getopt ("ab:c;d");
    $opt_a		and print "Option a is set\n";
    $opt_b =~ m/\S/	and print "Option b is set to $opt_b\n";
    defined $opt_c	and print "Option c is set",
			      $opt_c eq "1" ? "" : " to $opt_c", "\n";

=head1 DESCRIPTION

=over 4

=item  

=item B<getopt ($list);>

Set optional program flags as option letter prefixed with B<$opt_>.
Boolean options are set to B<1> (default undef).
Letters suffixed with a I<colon> (:) are set to the B<argument>, optionaly
separated with space.
Letters suffixed with a I<> or I<|> are pushed with the B<argument>,
optionaly separated with space. With I<|>, all arguments are prefixed with
the original option, with I<> it's stripped.
Letters suffixed with a I<semi-colon> (;) are set to the B<argument>,
optionaly separated with space, if passed, otherwise to B<1>.

B<getopt ()> returns the I<reorganized list of options parsed>. For above
example C<"-abzzz -c"> will return C<"-a -bzzz -c">. This string can be
used to post-evaluate the options parsed. In B<PROCURA::getopt ()>, it
is impossible to enable options like C<"-12">. B<getopt ("0123456789")>
will return C<"-1 -2">, which can be used to rebuild the effect of the
option passed.

I<$opt_V> is used as debug flag.

If B<getopt ()> fails to validate options/argumants, it calls B<usage ()> if
defined, otherwise, it prints an error message and exits with status 0.
B<usage ()> should explicitly terminite the program if wanted.

B<getopt ()> knows about the options C<-?>, C<-help> or C<--help>. If any
of these options is passed as a I<single argument> in C<@ARGV>, B<getopt ()>
will call B<usage ()> if defined.

=back

=cut

### ###########################################################################

use Exporter;
use vars qw(@ISA @EXPORT);

@ISA	 = qw(Exporter);
@EXPORT	 = qw(getopt); 

use strict;
no  strict "refs";

use vars qw (*opt_v *opt_V);
*opt_v = \$main::opt_v;
*opt_V = \$main::opt_V;

sub getopt
{
    my $list = shift;

    $] >= 5.008 and print STDERR
	"Waarschuwing: PROCURA::getopt is niet meer in onderhoud\n",
	"              gebruik Getopt::Long\n\a";

    sub my_usage
    {
	my $list = shift;
	defined &usage and return usage ();
	print STDERR "$0 did not define usage () for known option '$list', sorry\n";
	exit 0;
	} # my_usage

    @ARGV == 1 and $ARGV[0] eq "-?" || $ARGV[0] =~ m/^-+help$/ and my_usage ($list);

    my @opt = ();
    my $arg;

    my @opts = split m/ */, $list;
    while (@ARGV && ($arg = $ARGV[0]) =~ m/^-(.)(.*)/ && $arg ne "--") {
	my ($opt, $rest) = ($1, $2);

	$opt =~ m/[:;|\247]/ and print STDERR "'$opt' is not allowed as option\n", next;
	my $idx = index $list, $opt;

	if ($idx < $[) {
	    print STDERR "Unknown option: $opt\n";
	    my_usage ($list);
	    }

	shift @ARGV;
	my $Opt = "opt_$opt";
	if ($#opts > $idx && $opts[$idx + 1] eq ":") {
	    if ($rest eq "") {
		unless ($rest = shift @ARGV) {
		    print STDERR "Option $opt requires argument\n";
		    my_usage ($list);
		    }
		}
	    $opt_V and print STDERR "Option $opt set to '$rest'\n";
	    $$Opt = $rest;
	    push @opt, "-$opt$rest";
	    next;
	    }

	if ($#opts > $idx && $opts[$idx + 1] =~ m/([|])/) {
	    if ($rest eq "") {
		unless ($rest = shift @ARGV) {
		    print STDERR "Option $opt requires argument\n";
		    my_usage ($list);
		    }
		}
	    $opt_V and print STDERR "Option $opt extended with '$rest'\n";
	    $1 eq "" and push @$Opt, $rest or push @$Opt, "-$opt$rest";
	    push @opt, "-$opt$rest";
	    next;
	    }

	# This is an extension: optional argument
	# the argument cannot be separated by whitespace
	my $yn_val = 1;
	if ($#opts > $idx && $opts[$idx + 1] eq ";") {
	    if ($opt eq "v") {
		$rest =~ s/^(\d+)// and $yn_val = $1 + 0;
		}
	    elsif ($rest ne "") {
		$opt_V and print STDERR "Option $Opt set to '$rest'\n";
		$$Opt = $rest;
		push @opt, "-$opt$rest";
		next;
		}
	    }

	$opt eq "s" and ($opt_v = 0, $opt_V = 0);

	$opt_V and print STDERR "Option $opt set to $yn_val\n";
	$$Opt = $yn_val;
	push @opt, $yn_val == 1 ? "-$opt" : "-$opt$yn_val";

	$rest ne "" and unshift @ARGV, "-$rest";
	}
    my $opts = join " ", @opt;
    $opt_V and print STDERR "Reorganized options parsed: $opts\n";
    $opts;
    } # getopts

1;
